# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import html
import io
import json
import logging
import os
import shutil
import sys

from collections import defaultdict
from string import Template
from typing import Callable, Dict, List, Optional, Set, Tuple

from codechecker_report_converter.report import BugPathEvent, \
    InvalidFileContentMsg, File, MacroExpansion, Report, report_file, \
    reports as reports_helper
from codechecker_report_converter.report.statistics import Statistics
from codechecker_report_converter.report.checker_labels import CheckerLabels

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=no-name-in-module
else:
    from mypy_extensions import TypedDict


LOG = logging.getLogger('report-converter')


SkipReportHandler = Callable[
    [str, str, int, str, dict, Dict[int, str]],
    Tuple[bool, list]
]


class HTMLBugPathEvent(TypedDict):
    message: str
    fileId: str
    line: int
    column: int


HTMLBugPathEvents = List[HTMLBugPathEvent]


class HTMLMacroExpansion(HTMLBugPathEvent):
    name: str


HTMLMacroExpansions = List[HTMLMacroExpansion]


class Checker(TypedDict):
    name: str
    url: Optional[str]


class HTMLReport(TypedDict):
    fileId: str
    reportHash: Optional[str]
    checker: Checker
    analyzerName: Optional[str]
    line: int
    column: int
    message: str
    events: HTMLBugPathEvents
    macros: HTMLMacroExpansions
    notes: HTMLBugPathEvents
    reviewStatus: Optional[str]
    severity: Optional[str]
    testcase: Optional[str]
    timestamp: Optional[str]


HTMLReports = List[HTMLReport]


class FileSource(TypedDict):
    id: str
    filePath: str
    content: str


FileSources = Dict[str, FileSource]


class HtmlReportLink(TypedDict):
    report: HTMLReport
    link: str


def get_file_content(file_path: str) -> str:
    """ Return file content of the given file. """
    with open(file_path, 'r', encoding='utf-8', errors='replace') as f:
        return f.read()


class HtmlBuilder:
    """
    Helper class to create html file from a report data.
    """
    def __init__(
        self,
        layout_dir: str,
        checker_labels: Optional[CheckerLabels] = None
    ):
        self._checker_labels = checker_labels
        self.layout_dir = layout_dir
        self.generated_html_reports: Dict[str, HTMLReports] = {}
        self.files: FileSources = {}

        css_dir = os.path.join(self.layout_dir, 'css')
        js_dir = os.path.join(self.layout_dir, 'js')
        codemirror_dir = os.path.join(
            self.layout_dir, 'vendor', 'codemirror')

        # Mapping layout tags to files.
        self._layout_tag_files = {
            'style_css': os.path.join(css_dir, 'style.css'),
            'buglist_css': os.path.join(css_dir, 'buglist.css'),
            'bugview_css': os.path.join(css_dir, 'bugview.css'),
            'statistics_css': os.path.join(css_dir, 'statistics.css'),
            'icon_css': os.path.join(css_dir, 'icon.css'),
            'table_css': os.path.join(css_dir, 'table.css'),
            'codemirror_license': os.path.join(codemirror_dir,
                                               'codemirror.LICENSE'),
            'codemirror_css': os.path.join(codemirror_dir,
                                           'codemirror.min.css'),
            'codemirror_js': os.path.join(codemirror_dir, 'codemirror.min.js'),
            'clike_js': os.path.join(codemirror_dir, 'clike.min.js'),
            'bug_viewer': os.path.join(js_dir, 'bugviewer.js'),
            'bug_list': os.path.join(js_dir, 'buglist.js'),
            'browser_support': os.path.join(js_dir, 'browsersupport.js')
        }

        # Get the HTML layout file content.
        self._layout = Template(get_file_content(
            os.path.join(self.layout_dir, 'layout.html')))

        self._index = Template(get_file_content(
            os.path.join(self.layout_dir, 'index.html')))

        self._statistics = Template(get_file_content(
            os.path.join(self.layout_dir, 'statistics.html')))

        # Get the content of the HTML layout dependencies.
        self._tag_contents = {}
        for tag in self._layout_tag_files:
            self._tag_contents[tag] = get_file_content(
                self._layout_tag_files[tag])

    def get_severity(self, checker_name: str) -> str:
        """ Returns severity level for the given checker name. """
        return self._checker_labels.severity(checker_name) \
            if self._checker_labels else 'UNSPECIFIED'

    def _add_source_file(self, file: File) -> FileSource:
        """
        Updates file source data by file id if the given file hasn't been
        processed.
        """
        if file.id in self.files:
            return self.files[file.id]

        try:
            file_content = file.content
        except Exception:
            file_content = InvalidFileContentMsg

        self.files[file.id] = {
            'id': file.id, 'filePath': file.path,
            'content': html.escape(file_content)
        }

        return self.files[file.id]

    def _get_doc_url(self, report: Report) -> Optional[str]:
        """ Get documentation url for the given report if exists. """
        if self._checker_labels:
            doc_urls = self._checker_labels.label_of_checker(
                report.checker_name, 'doc_url', report.analyzer_name)
            return doc_urls[0] if doc_urls else None

        return None

    def _get_html_reports(
        self,
        reports: List[Report]
    ) -> Tuple[HTMLReports, FileSources]:
        """ Get HTML reports from the given reports.

        Returns a list of html reports and references to file sources.
        """
        html_reports: HTMLReports = []
        files: FileSources = {}

        def to_bug_path_events(
            events: List[BugPathEvent]
        ) -> HTMLBugPathEvents:
            """ Converts the given events to html compatible format. """
            html_events: HTMLBugPathEvents = []
            for event in events:
                files[event.file.id] = self._add_source_file(event.file)

                html_events.append({
                    'message': event.message,
                    'fileId': event.file.id,
                    'line': event.line,
                    'column': event.column,
                })
            return html_events

        def to_macro_expansions(
            macro_expansions: List[MacroExpansion]
        ) -> HTMLMacroExpansions:
            """ Converts the given events to html compatible format. """
            html_macro_expansions: HTMLMacroExpansions = []
            for macro_expansion in macro_expansions:
                files[macro_expansion.file.id] = self._add_source_file(
                    macro_expansion.file)

                html_macro_expansions.append({
                    'message': macro_expansion.message,
                    'name': macro_expansion.name,
                    'fileId': macro_expansion.file.id,
                    'line': macro_expansion.line,
                    'column': macro_expansion.column,
                })
            return html_macro_expansions

        for report in reports:
            files[report.file.id] = self._add_source_file(report.file)

            html_reports.append({
                'fileId': report.file.id,
                'reportHash': report.report_hash,
                'checker': {
                    'name': report.checker_name,
                    'url': self._get_doc_url(report) or ''
                },
                'analyzerName': report.analyzer_name,
                'line': report.line,
                'column': report.column,
                'message': report.message,
                'events': to_bug_path_events(report.bug_path_events),
                'macros': to_macro_expansions(report.macro_expansions),
                'notes': to_bug_path_events(report.notes),
                'reviewStatus': report.review_status.formatted_status()
                if report.review_status else '',
                'severity': self.get_severity(report.checker_name),
                'testcase': report.annotations.get('testcase')
                if report.annotations else None,
                'timestamp': report.annotations.get('timestamp')
                if report.annotations else None
            })

        return html_reports, files

    def create(
        self,
        output_file_path: str,
        reports: List[Report]
    ) -> Tuple[Optional[HTMLReports], Set[str]]:
        """
        Create html file from the given analyzer result file to the output
        path.
        """
        changed_files = reports_helper.get_changed_files(reports)

        if changed_files:
            return None, changed_files

        html_reports, files = self._get_html_reports(reports)

        self.generated_html_reports[output_file_path] = html_reports

        substitute_data = self._tag_contents
        substitute_data.update({
            'report_data': json.dumps({
                'files': files,
                'reports': html_reports
            })
        })

        content = self._layout.substitute(substitute_data)

        with open(output_file_path, 'w+',
                  encoding='utf-8', errors='replace') as f:
            f.write(content)

        return html_reports, changed_files

    def create_index_html(self, output_dir: str):
        """
        Creates an index.html file which lists all available bugs which was
        found in the processed plist files. This also creates a link for each
        bug to the created html file where the bug can be found.
        """
        # Sort reports based on file path levels.
        html_report_links: List[HtmlReportLink] = []
        for html_file, reports in self.generated_html_reports.items():
            for report in reports:
                html_report_links.append({'link': html_file, 'report': report})

        table_reports = map(lambda data: {
            'link': os.path.basename(data['link']),
            'file-path': data['report']['fileId'],
            'report-hash': data['report']['reportHash'],
            'checker-name': data['report']['checker']['name'],
            'checker-url': data['report']['checker']['url'],
            'line': data['report']['line'],
            'message': data['report']['message'],
            'review-status': data['report']['reviewStatus'],
            'severity': data['report']['severity'],
            'bug-path-length': len(data['report']['events']),
            'testcase': data['report']['testcase'],
            'timestamp': data['report']['timestamp']
        }, html_report_links)

        self._tag_contents['table_reports'] = json.dumps(list(table_reports))

        content = self._index.substitute(self._tag_contents)
        output_path = os.path.join(output_dir, 'index.html')
        with open(output_path, 'w+', encoding='utf-8',
                  errors='replace') as html_output:
            html_output.write(content)

    def create_statistics_html(self, output_dir: str):
        """
        Creates a statistics.html file which contains statistics information
        from the HTML generation process.
        """
        def severity_order(severity: str) -> int:
            """
            This function determines in which order severities should be
            printed to the output. This function can be given via "key"
            attribute to sort() function.
            """
            severities = ['CRITICAL', 'HIGH', 'MEDIUM', 'LOW', 'STYLE',
                          'UNSPECIFIED']
            return severities.index(severity)

        num_of_analyzer_result_files = len(self.generated_html_reports)

        num_of_reports = 0
        for html_file in self.generated_html_reports:
            num_of_reports += len(self.generated_html_reports[html_file])

        checker_statistics: Dict[str, int] = defaultdict(int)
        for html_file in self.generated_html_reports:
            for report in self.generated_html_reports[html_file]:
                checker = report['checker']['name']
                checker_statistics[checker] += 1

        checker_rows: List[List[str]] = []
        severity_statistics: Dict[str, int] = defaultdict(int)

        with io.StringIO() as string:
            for checker_name in sorted(checker_statistics):
                severity = self.get_severity(checker_name)
                string.write('''
                  <tr>
                    <td>{0}</td>
                    <td class="severity" severity="{1}">
                      <i class="severity-{1}" title="{1}"></i>
                    </td>
                    <td>{2}</td>
                  </tr>
                '''.format(checker_name, severity.lower(),
                           checker_statistics[checker_name]))
                checker_rows.append([checker_name, severity,
                                    str(checker_statistics[checker_name])])
                severity_statistics[severity] += \
                    checker_statistics[checker_name]
            checker_statistics_content = string.getvalue()

        severity_rows: List[List[str]] = []

        with io.StringIO() as string:
            for severity in sorted(severity_statistics, key=severity_order):
                num = severity_statistics[severity]
                string.write('''
                  <tr>
                    <td class="severity" severity="{0}">
                      <i class="severity-{0}" title="{0}"></i>
                    </td>
                    <td>{1}</td>
                  </tr>
                '''.format(severity.lower(), num))
                severity_rows.append([severity, str(num)])
            severity_statistics_content = string.getvalue()

        substitute_data = self._tag_contents
        substitute_data.update({
            'num_of_analyzer_result_files': str(num_of_analyzer_result_files),
            'number_of_reports': str(num_of_reports),
            'checker_statistics': checker_statistics_content,
            'severity_statistics': severity_statistics_content})

        content = self._statistics.substitute(substitute_data)

        output_path = os.path.join(output_dir, 'statistics.html')
        with open(output_path, 'w+', encoding='utf-8',
                  errors='ignore') as html_output:
            html_output.write(content)

    def finish(self, output_dir_path: str, statistics: Statistics):
        """ Creates common html files and print summary messages. """
        self.create_index_html(output_dir_path)
        self.create_statistics_html(output_dir_path)
        statistics.write()

        print(f"\nTo view statistics in a browser run:\n> firefox "
              f"{os.path.join(output_dir_path, 'statistics.html')}")

        print(f"\nTo view the results in a browser run:\n> firefox "
              f"{os.path.join(output_dir_path, 'index.html')}")


def convert(
    file_path: str,
    reports: List[Report],
    output_dir_path: str,
    html_builder: HtmlBuilder
) -> Set[str]:
    """
    Prints the results in the given file to HTML file.

    Returns the skipped analyzer result files because of source
    file content change.
    """
    if not reports:
        LOG.info(f'No report data in {file_path} file.')
        return set()

    html_filename = f"{os.path.basename(file_path)}.html"
    html_output_path = os.path.join(output_dir_path, html_filename)
    _, changed_files = html_builder.create(
        html_output_path, reports)

    if changed_files:
        return changed_files

    LOG.info(f"Html file was generated: {html_output_path}")
    return changed_files


def parse(
    input_path: str,
    output_path: str,
    layout_dir: str,
    html_builder: Optional[HtmlBuilder] = None
) -> Set[str]:
    """
    Parses analyzer result files from the given input directory to the output
    directory.

    Return a set of changed files.
    """
    files = []
    input_path = os.path.abspath(input_path)
    output_dir = os.path.abspath(output_path)

    if os.path.exists(output_path):
        LOG.info("Previous analysis results in '%s' have been removed, "
                 "overwriting with current results.", output_dir)
        shutil.rmtree(output_path)

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    if os.path.isfile(input_path):
        files.append(input_path)
    elif os.path.isdir(input_path):
        _, _, file_names = next(os.walk(input_path), ([], [], []))
        files = [os.path.join(input_path, file_name) for file_name
                 in file_names]

    # Source files which modification time changed since the last analysis.
    changed_source_files: Set[str] = set()

    if not html_builder:
        html_builder = HtmlBuilder(layout_dir)

    for file_path in files:
        if not report_file.is_supported(file_path):
            LOG.info("\nSkipping input file %s as it is not supported "
                     "analyzer result file.", file_path)
            continue

        LOG.info(f"\nParsing input file '%s'", file_path)

        reports = report_file.get_reports(file_path)
        changed_source = convert(file_path, reports, output_path, html_builder)

        if changed_source:
            changed_source_files.update(changed_source)

    return changed_source_files



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

from typing import List

from codechecker_report_converter.report import Report

from ..analyzer_result import AnalyzerResultBase
from .parser import Parser


class AnalyzerResult(AnalyzerResultBase):
    """ Transform analyzer result of Sparse. """

    TOOL_NAME = 'sparse'
    NAME = 'Sparse'
    URL = 'https://git.kernel.org/pub/scm/devel/sparse/sparse.git'

    def get_reports(self, file_path: str) -> List[Report]:
        """ Get reports from the given analyzer result. """
        return Parser(file_path).get_reports(file_path)



"""
Initial schema

Revision ID: 82ca43f05c10
Revises:     <None>
Create Date: 2017-09-18 21:00:11.593693
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = '82ca43f05c10'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'comments',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('bug_hash', sa.String(), nullable=False),
        sa.Column('author', sa.String(), nullable=False),
        sa.Column('message', sa.Binary(), nullable=False),
        sa.Column('created_at', sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_comments'))
    )
    op.create_index(op.f('ix_comments_bug_hash'), 'comments', ['bug_hash'],
                    unique=False)

    op.create_table(
        'db_version',
        sa.Column('major', sa.Integer(), nullable=False),
        sa.Column('minor', sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint('major', 'minor', name=op.f('pk_db_version'))
    )

    op.create_table(
        'file_contents',
        sa.Column('content_hash', sa.String(), nullable=False),
        sa.Column('content', sa.Binary(), nullable=True),
        sa.PrimaryKeyConstraint('content_hash', name=op.f('pk_file_contents'))
    )

    op.create_table(
        'review_statuses',
        sa.Column('bug_hash', sa.String(), nullable=False),
        sa.Column('status', sa.Enum('unreviewed',
                                    'confirmed',
                                    'false_positive',
                                    'intentional',
                                    name='review_status'),
                  nullable=False),
        sa.Column('author', sa.String(), nullable=False),
        sa.Column('message', sa.Binary(), nullable=False),
        sa.Column('date', sa.DateTime(), nullable=False),
        sa.PrimaryKeyConstraint('bug_hash', name=op.f('pk_review_statuses'))
    )

    op.create_table(
        'runs',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('date', sa.DateTime(), nullable=True),
        sa.Column('duration', sa.Integer(), nullable=True),
        sa.Column('name', sa.String(), nullable=True),
        sa.Column('version', sa.String(), nullable=True),
        sa.Column('command', sa.String(), nullable=True),
        sa.Column('can_delete', sa.Boolean(), server_default=sa.text('true'),
                  nullable=False),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_runs')),
        sa.UniqueConstraint('name', name=op.f('uq_runs_name'))
    )

    op.create_table(
        'files',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('filepath', sa.String(), nullable=False),
        sa.Column('filename', sa.String(), nullable=False),
        sa.Column('content_hash', sa.String(), nullable=True),
        sa.ForeignKeyConstraint(['content_hash'],
                                ['file_contents.content_hash'],
                                name=op.f(
                                    'fk_files_content_hash_file_contents'),
                                ondelete='CASCADE', initially='DEFERRED',
                                deferrable=True),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_files')),
        sa.UniqueConstraint('filepath', 'content_hash',
                            name=op.f('uq_files_filepath'))
    )

    op.create_table(
        'run_histories',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('run_id', sa.Integer(), nullable=True),
        sa.Column('version_tag', sa.String(), nullable=True),
        sa.Column('user', sa.String(), nullable=False),
        sa.Column('time', sa.DateTime(), nullable=False),
        sa.ForeignKeyConstraint(['run_id'], ['runs.id'],
                                name=op.f('fk_run_histories_run_id_runs'),
                                ondelete='CASCADE', initially='DEFERRED',
                                deferrable=True),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_run_histories')),
        sa.UniqueConstraint('run_id', 'version_tag',
                            name=op.f('uq_run_histories_run_id'))
    )
    op.create_index(op.f('ix_run_histories_run_id'), 'run_histories',
                    ['run_id'], unique=False)

    op.create_table(
        'reports',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('file_id', sa.Integer(), nullable=True),
        sa.Column('run_id', sa.Integer(), nullable=True),
        sa.Column('bug_id', sa.String(), nullable=True),
        sa.Column('checker_id', sa.String(), nullable=True),
        sa.Column('checker_cat', sa.String(), nullable=True),
        sa.Column('bug_type', sa.String(), nullable=True),
        sa.Column('severity', sa.Integer(), nullable=True),
        sa.Column('line', sa.Integer(), nullable=True),
        sa.Column('column', sa.Integer(), nullable=True),
        sa.Column('checker_message', sa.String(), nullable=True),
        sa.Column('detection_status', sa.Enum('new',
                                              'unresolved',
                                              'resolved',
                                              'reopened',
                                              name='detection_status'),
                  nullable=True),
        sa.Column('detected_at', sa.DateTime(), nullable=False),
        sa.Column('fixed_at', sa.DateTime(), nullable=True),
        sa.ForeignKeyConstraint(['file_id'], ['files.id'],
                                name=op.f('fk_reports_file_id_files'),
                                ondelete='CASCADE', initially='DEFERRED',
                                deferrable=True),
        sa.ForeignKeyConstraint(['run_id'], ['runs.id'],
                                name=op.f('fk_reports_run_id_runs'),
                                ondelete='CASCADE', initially='DEFERRED',
                                deferrable=True),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_reports'))
    )
    op.create_index(op.f('ix_reports_bug_id'), 'reports', ['bug_id'],
                    unique=False)
    op.create_index(op.f('ix_reports_run_id'), 'reports', ['run_id'],
                    unique=False)

    op.create_table(
        'bug_path_events',
        sa.Column('line_begin', sa.Integer(), nullable=True),
        sa.Column('col_begin', sa.Integer(), nullable=True),
        sa.Column('line_end', sa.Integer(), nullable=True),
        sa.Column('col_end', sa.Integer(), nullable=True),
        sa.Column('order', sa.Integer(), nullable=False),
        sa.Column('msg', sa.String(), nullable=True),
        sa.Column('file_id', sa.Integer(), nullable=True),
        sa.Column('report_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['file_id'], ['files.id'],
                                name=op.f('fk_bug_path_events_file_id_files'),
                                ondelete='CASCADE', initially='DEFERRED',
                                deferrable=True),
        sa.ForeignKeyConstraint(['report_id'], ['reports.id'],
                                name=op.f(
                                    'fk_bug_path_events_report_id_reports'),
                                ondelete='CASCADE', initially='DEFERRED',
                                deferrable=True),
        sa.PrimaryKeyConstraint('order', 'report_id',
                                name=op.f('pk_bug_path_events'))
    )
    op.create_index(op.f('ix_bug_path_events_file_id'), 'bug_path_events',
                    ['file_id'], unique=False)

    op.create_table(
        'bug_report_points',
        sa.Column('line_begin', sa.Integer(), nullable=True),
        sa.Column('col_begin', sa.Integer(), nullable=True),
        sa.Column('line_end', sa.Integer(), nullable=True),
        sa.Column('col_end', sa.Integer(), nullable=True),
        sa.Column('order', sa.Integer(), nullable=False),
        sa.Column('file_id', sa.Integer(), nullable=True),
        sa.Column('report_id', sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(['file_id'], ['files.id'],
                                name=op.f(
                                    'fk_bug_report_points_file_id_files'),
                                ondelete='CASCADE', initially='DEFERRED',
                                deferrable=True),
        sa.ForeignKeyConstraint(
            ['report_id'], ['reports.id'],
            name=op.f('fk_bug_report_points_report_id_reports'),
            ondelete='CASCADE', initially='DEFERRED', deferrable=True),
        sa.PrimaryKeyConstraint('order', 'report_id',
                                name=op.f('pk_bug_report_points'))
    )
    op.create_index(op.f('ix_bug_report_points_file_id'), 'bug_report_points',
                    ['file_id'], unique=False)


def downgrade():
    op.drop_index(op.f('ix_bug_report_points_file_id'),
                  table_name='bug_report_points')
    op.drop_table('bug_report_points')
    op.drop_index(op.f('ix_bug_path_events_file_id'),
                  table_name='bug_path_events')
    op.drop_table('bug_path_events')
    op.drop_index(op.f('ix_reports_run_id'),
                  table_name='reports')
    op.drop_index(op.f('ix_reports_bug_id'),
                  table_name='reports')
    op.drop_table('reports')
    op.drop_index(op.f('ix_run_histories_run_id'),
                  table_name='run_histories')
    op.drop_table('run_histories')
    op.drop_table('files')
    op.drop_table('runs')
    op.drop_table('review_statuses')
    op.drop_table('file_contents')
    op.drop_table('db_version')
    op.drop_index(op.f('ix_comments_bug_hash'), table_name='comments')
    op.drop_table('comments')



"""
Content hash index for files

Revision ID: 2185167f8568
Revises:     5f8a443a51e5
Create Date: 2020-07-28 16:02:01.131126
"""

from alembic import op

# Revision identifiers, used by Alembic.
revision = '2185167f8568'
down_revision = '5f8a443a51e5'
branch_labels = None
depends_on = None


def upgrade():
    op.create_index(op.f('ix_files_content_hash'), 'files', ['content_hash'],
                    unique=False)


def downgrade():
    op.drop_index(op.f('ix_files_content_hash'), table_name='files')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""Argument parsing related helper classes and functions."""


import argparse
import textwrap


class RawDescriptionDefaultHelpFormatter(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter
):
    """
    Adds default values to argument help and retains any formatting in
    descriptions.
    """
    def _split_lines(self, text, width):
        """ Split the lines.

        If the text parameter starts with 'R|' it will keep whitespaces and
        it will wrapp the content. Otherwise it will call the parent function
        of RawDescriptionHelpFormatter.
        """
        if text.startswith('R|'):
            lines = [textwrap.wrap(line, width, replace_whitespace=False)
                     for line in text[2:].lstrip().splitlines()]
            return [line for sublines in lines for line in sublines]

        return argparse.RawDescriptionHelpFormatter._split_lines(self, text,
                                                                 width)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Result handler for Cppcheck.
"""
from typing import Optional

from codechecker_report_converter.report.parser.base import AnalyzerInfo
from codechecker_report_converter.analyzers.cppcheck.analyzer_result import \
    AnalyzerResult
from codechecker_report_converter.report import BugPathEvent, \
        Range, report_file
from codechecker_report_converter.report.hash import get_report_hash, HashType

from codechecker_common.logger import get_logger
from codechecker_common.skiplist_handler import SkipListHandlers
from codechecker_common.review_status_handler import ReviewStatusHandler

from ..result_handler_base import ResultHandler

LOG = get_logger('analyzer.cppcheck')


class CppcheckResultHandler(ResultHandler):
    """
    Create analyzer result file for Cppcheck output.
    """

    def __init__(self, *args, **kwargs):
        self.analyzer_info = AnalyzerInfo(name=AnalyzerResult.TOOL_NAME)

        super(CppcheckResultHandler, self).__init__(*args, **kwargs)

    def postprocess_result(
        self,
        skip_handlers: Optional[SkipListHandlers],
        rs_handler: Optional[ReviewStatusHandler]
    ):
        """
        Generate analyzer result output file which can be parsed and stored
        into the database.
        """
        LOG.debug_analyzer(self.analyzer_stdout)

        reports = report_file.get_reports(
            self.analyzer_result_file, self.checker_labels,
            source_dir_path=self.source_dir_path)

        reports = [r for r in reports if not r.skip(skip_handlers)]

        for report in reports:
            # TODO check if prefix cascading still occurs.
            if not report.checker_name.startswith("cppcheck-"):
                report.checker_name = "cppcheck-" + report.checker_name

        hash_type = HashType.PATH_SENSITIVE
        if self.report_hash_type == 'context-free-v2':
            hash_type = HashType.CONTEXT_FREE
        elif self.report_hash_type == 'diagnostic-message':
            hash_type = HashType.DIAGNOSTIC_MESSAGE

        for report in reports:
            report.report_hash = get_report_hash(report, hash_type)
            bpe = BugPathEvent(
                    report.message,
                    report.file,
                    report.line,
                    report.column,
                    Range(report.line,
                          report.column,
                          report.line,
                          report.column))
            if bpe != report.bug_path_events[-1]:
                report.bug_path_events.append(bpe)

        if rs_handler:
            reports = [r for r in reports if not rs_handler.should_ignore(r)]

        report_file.create(
            self.analyzer_result_file, reports, self.checker_labels,
            self.analyzer_info)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import json
import logging
import os

from sarif import loader

from typing import Dict, List, Optional

from urllib.parse import urlparse
from typing import Any, List, Optional, Tuple

from codechecker_report_converter.report import BugPathEvent, \
    BugPathPosition, File, MacroExpansion, get_or_create_file, Range, Report
from codechecker_report_converter.report.hash import get_report_hash, HashType
from codechecker_report_converter.report.parser.base import AnalyzerInfo, \
    BaseParser, load_json, get_tool_info


EXTENSION = 'sarif'


LOG = logging.getLogger('report-converter')


# §3.37
class ThreadFlowInfo:
    def __init__(self):
        self.bug_path_events: List[BugPathEvent] = []
        self.bug_path_positions: List[BugPathPosition] = []
        self.notes: List[BugPathEvent] = []
        self.macro_expansions: List[MacroExpansion] = []


# Parse to/from sarif formats.
# Sarif has an extensive documentation, and supports a grand number of fields
# that CodeChecker has no use for as of writing this comment (e.g. §3.39 graph,
# §3.30 region), leading to these fields to be ignored.
# This is NOT (and will likely never be) a full fledged sarif parser, only what
# we need from it at the time.
class Parser(BaseParser):
    def get_reports(
        self,
        result_file_path: str,
        source_dir_path: Optional[str] = None
    ) -> List[Report]:
        """ Get reports from the given analyzer result file. """

        if not self.has_any_runs(result_file_path):
            return []

        reports: List[Report] = []
        self.result_file_path = result_file_path
        self.had_error = False

        data = loader.load_sarif_file(result_file_path)

        for run in data.runs:
            rules = self._get_rules(run.run_data)
            # $3.14.14
            self.original_uri_base_ids = None
            if "originalUriBaseIds" in run.run_data:
                self.original_uri_base_ids = run.run_data["originalUriBaseIds"]

            if "results" not in run.run_data:
                continue

            for result in run.get_results():
                rule_id = result["ruleId"]
                message = self._process_message(
                    result["message"], rule_id, rules)  # §3.11

                thread_flow_info = self._process_code_flows(
                    result, rule_id, rules)
                for location in result.get("locations", []):
                    # TODO: We don't really support non-local analyses, so we
                    # only parse physical locations here.
                    file, rng = self._process_location(location)
                    if not (file and rng):
                        continue
                    if self.had_error:
                        return []

                    bug_path_events = thread_flow_info.bug_path_events or None

                    report = Report(
                        file, rng.start_line, rng.start_col,
                        message, rule_id,  # TODO: Add severity.
                        analyzer_result_file_path=result_file_path,
                        bug_path_events=bug_path_events,
                        bug_path_positions=thread_flow_info.bug_path_positions,
                        notes=thread_flow_info.notes,
                        macro_expansions=thread_flow_info.macro_expansions)

                    if report.report_hash is None:
                        report.report_hash = get_report_hash(
                            report, HashType.PATH_SENSITIVE)

                    reports.append(report)

        return reports

    def has_any_runs(self, result_file_path: str) -> bool:
        """
        The package that contains the sarif import, sarif-tools currently
        crashes when a sarif file contains no runs, so we need to manually
        check it.
        """
        ret = load_json(result_file_path)
        return bool(ret and ret["runs"])

    def _get_rules(self, data: Dict) -> Dict[str, Dict]:
        """
        The list of checkers as per §3.19.23. The standard doesn't
        explicitly state whether this is the list of available checkers, or
        enabled checkers, unfortunately.
        """
        rules: Dict[str, Dict] = {}

        driver = data["tool"]["driver"]
        for rule in driver.get("rules", []):
            rules[rule["id"]] = rule

        return rules

    def _process_code_flows(
        self,
        result: Dict,
        rule_id: str,
        rules: Dict[str, Dict]
    ) -> ThreadFlowInfo:
        """
        Parse threadFlows (§3.36.3), which is a list of threadFlow objects
        (§3.37), basically the components of a bug path (events, notes, arrows,
        etc).
        """

        thread_flow_info = ThreadFlowInfo()

        # TODO: Currently, we only collect bug path events.

        for code_flow in result.get("codeFlows", []):
            for thread_flow in code_flow.get("threadFlows", []):
                for location_data in thread_flow["locations"]:
                    # There are a lot data stored alongside the location worth
                    # parsing, but we only need the actual location now.
                    location = location_data["location"]

                    if "message" not in location:
                        # TODO: This might be a bug path position (for arrows).
                        continue

                    message = self._process_message(
                        location["message"], rule_id, rules)

                    file, rng = self._process_location(location)
                    if not (file and rng):
                        continue

                    thread_flow_info.bug_path_events.append(BugPathEvent(
                        message, file, rng.start_line, rng.start_col, rng))

        return thread_flow_info

    def _process_location(
        self,
        location: Dict,
    ) -> Tuple[Optional[File], Optional[Range]]:
        """
        Parse location (§3.28). Currently we only parse physical locations
        (§3.29), that describe a (file, range) pair, among other things.
        Sarif also supports logical locations (§3.33) that might describe
        a namespace, type, etc, but CodeChecker has no use for that.
        """
        physical_loc = location.get("physicalLocation")
        if physical_loc:
            file = self._get_file(physical_loc)
            rng = self._get_range(physical_loc)
            return file, rng

        return None, None

    def _get_range(self, physical_loc: Dict) -> Optional[Range]:
        """ Get range from a physical location. """
        region = physical_loc.get("region", {})
        start_line = region.get("startLine")
        if start_line is None:
            return None

        start_col = region.get("startColumn", 1)
        end_line = region.get("endLine", start_line)
        end_col = region.get("endColumn", start_col)

        return Range(start_line, start_col, end_line, end_col)

    def _resolve_uri_base_id(self, uri_base_id: str, postfix: str):
        """
        Try to assemble the leading part of the URI needed to find out the
        absolute path of the file. If this fails, we consider the entire sarif
        file invalid, not only this particular record.
        """
        error_str = \
            f"Failed to fully resolve the path for '{postfix}'. sarif " \
            "v2.1.0 permits this, but CodeChecker requires it to be " \
            "available. If you intend to hide the absolute path before " \
            "storing the results to a CodeChecker server, use " \
            "--trim-path-prefix."

        if not self.original_uri_base_ids:
            LOG.error("Missing 'originalUriBaseIds' (sarif v2.1.0 §3.14.14) "
                      f"in '{self.result_file_path}'.")
            LOG.error(error_str)
            self.had_error = True
            return ""

        original = self.original_uri_base_ids.get(uri_base_id)
        if not original:
            LOG.error(f"Missing entry for '{uri_base_id} in"
                      "'originalUriBaseIds' (sarif v2.1.0 §3.14.14)"
                      f"in '{self.result_file_path}'.")
            LOG.error(error_str)
            self.had_error = True
            return ""

        abs_uri_prefix = original.get("uri")
        if not abs_uri_prefix:
            LOG.warning("Missing 'uri' (sarif v2.1.0 §3.4.3) for "
                        f"'{uri_base_id} in 'originalUriBaseIds' (sarif "
                        f"v2.1.0 §3.14.14) in '{self.result_file_path}'.")
            LOG.error(error_str)
            self.had_error = True
            return ""

        if "uri_base_id" in original:
            prefix = self._resolve_uri_base_id(original.get("uri_base_id"),
                                               postfix)
            return prefix + abs_uri_prefix
        return abs_uri_prefix

    def _get_file(
        self,
        physical_loc: Dict
    ) -> Optional[File]:
        """
        Assemble the artifact location (§3.4) for physical_loc. Sarif files
        contain a lot of relative references, but also contains everything
        needed to resolve an absolute path for the file.
        """
        artifact_loc = physical_loc.get("artifactLocation")
        if not artifact_loc:
            return None

        uri = artifact_loc.get("uri")

        if "uriBaseId" in artifact_loc:
            uri = self._resolve_uri_base_id(
                    artifact_loc.get("uriBaseId"), uri) + uri

        uri_parsed = urlparse(uri)
        if uri_parsed is None:
            LOG.warning(f"Failed to urlparse {uri}!")
            return None

        file_path = os.path.join(uri_parsed.netloc, uri_parsed.path)

        return get_or_create_file(file_path, self._file_cache)

    def _process_message(
        self,
        msg: Dict,
        rule_id: str,
        rules: Dict[str, Dict]
    ) -> str:
        """
        Parse message (§3.11). Sometimes, the warning message is attributed to
        the checker (or rule, as is called here) under messageStrings
        (§3.49.11), which might need a little extra parsing.
        """
        if "text" in msg:
            return msg["text"]

        args = msg.get("arguments", [])

        rule = rules[rule_id]
        message_strings = rule.get("messageStrings", {})
        return message_strings[msg["id"]]["text"].format(*args)

    def convert(
        self,
        reports: List[Report],
        analyzer_info: Optional[AnalyzerInfo] = None
    ):
        """ Converts the given reports to sarif format. """
        tool_name, tool_version = get_tool_info()

        rules = {}
        results = []
        for report in reports:
            if report.checker_name not in rules:
                rules[report.checker_name] = {
                    "id": report.checker_name,
                    "fullDescription": {
                        "text": report.message
                    }
                }

            results.append(self._create_result(report))

        return {
            "version": "2.1.0",
            "$schema": "https://raw.githubusercontent.com/oasis-tcs/"
                       "sarif-spec/master/Schemata/sarif-schema-2.1.0.json",
            "runs": [{
                "tool": {
                    "driver": {
                        "name": tool_name,
                        "version": tool_version,
                        "rules": list(rules.values())
                    }
                },
                "results": results
            }]
        }

    def _create_result(self, report: Report) -> Dict:
        """ Create result dictionary from the given report. """
        result = {
            "ruleId": report.checker_name,
            "message": {
                "text": report.message
            },
            "locations": [{
                "physicalLocation": {
                    "artifactLocation": {
                        "uri": f"file://{report.file.original_path}"
                    },
                    "region": {
                        "startLine": report.line,
                        "startColumn": report.column
                    }
                }
            }]
        }

        locations = []

        if report.bug_path_events:
            for event in report.bug_path_events:
                locations.append(self._create_location_from_bug_path_event(
                    event, "important"))

        if report.notes:
            for note in report.notes:
                locations.append(self._create_location_from_bug_path_event(
                    note, "essential"))

        if report.macro_expansions:
            for macro_expansion in report.macro_expansions:
                locations.append(self._create_location_from_bug_path_event(
                    macro_expansion, "essential"))

        if report.bug_path_positions:
            for bug_path_position in report.bug_path_positions:
                locations.append(self._create_location(bug_path_position))

        if locations:
            result["codeFlows"] = [{
                "threadFlows": [{"locations": locations}]
            }]

        return result

    def _create_location_from_bug_path_event(
        self,
        event: BugPathEvent,
        importance: str
    ) -> Dict[str, Any]:
        """ Create location from bug path event. """
        location = self._create_location(event, event.line, event.column)

        location["importance"] = importance
        location["location"]["message"] = {"text": event.message}

        return location

    def _create_location(
        self,
        pos: BugPathPosition,
        line: Optional[int] = -1,
        column: Optional[int] = -1
    ) -> Dict[str, Any]:
        """ Create location from bug path position. """
        if pos.range:
            rng = pos.range
            region = {
                "startLine": rng.start_line,
                "startColumn": rng.start_col,
                "endLine": rng.end_line,
                "endColumn": rng.end_col,
            }
        else:
            # Seems like there is no smarter way to make mypy happy. Without
            # this trickery, it will complain about Optional[int] and int being
            # incompatible.
            assert line is not None
            assert column is not None
            linei: int = line
            columni: int = column
            region = {
                "startLine": linei,
                "startColumn": columni,
            }

        return {
            "location": {
                "physicalLocation": {
                    "artifactLocation": {
                        "uri": f"file://{pos.file.original_path}"
                    },
                    "region": region
                }
            }
        }

    def write(self, data: Any, output_file_path: str):
        """ Creates an analyzer output file from the given data. """
        try:
            with open(output_file_path, 'w',
                      encoding="utf-8", errors="ignore") as f:
                json.dump(data, f)
        except TypeError as err:
            LOG.error('Failed to write sarif file: %s', output_file_path)
            LOG.error(err)
            import traceback
            traceback.print_exc()

    def replace_report_hash(
        self,
        analyzer_result_file_path: str,
        hash_type=HashType.CONTEXT_FREE
    ):
        """
        Override hash in the given file by using the given version hash.
        """
        pass



"""
Run limitation for products

Revision ID: bb5278995f41
Revises:     150800b30447
Create Date: 2018-03-01 15:38:41.164141
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = 'bb5278995f41'
down_revision = '150800b30447'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('products',
                  sa.Column('run_limit', sa.Integer(), nullable=True))


def downgrade():
    op.drop_column('products', 'run_limit')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Module for handling ClangTidy-related functionality related to analysis,
checker handling and configuration.
"""

import ast
import json
import os
import re
import shlex
import subprocess
from typing import Iterable, List, Set, Tuple

import yaml

from codechecker_common.logger import get_logger

from codechecker_analyzer import analyzer_context, env
from codechecker_analyzer.analyzers.clangsa.analyzer import ClangSA

from .. import analyzer_base
from ..config_handler import CheckerState, CheckerType, \
    get_compiler_warning_name_and_type
from ..flag import has_flag
from ..flag import prepend_all

from . import config_handler
from . import result_handler

LOG = get_logger('analyzer')


def parse_checkers(tidy_output):
    """
    Parse clang tidy checkers list.
    Skip clang static analyzer checkers.
    Store them to checkers.
    """
    checkers = []
    pattern = re.compile(r'^\S+$')
    for line in tidy_output.splitlines():
        line = line.strip()
        if line.startswith('Enabled checks:') or line == '':
            continue
        elif line.startswith('clang-analyzer-'):
            continue
        match = pattern.match(line)
        if match:
            checkers.append((match.group(0), ''))
    return checkers


def parse_checker_config_old(config_dump):
    """
    Return the parsed clang-tidy config options as a list of
    (flag, default_value) tuples. This variant works
    for clang-tidy up to version 14.

    config_dump -- clang-tidy config options YAML dump in pre-LLVM15 format.
    """
    reg = re.compile(r'key:\s+(\S+)\s+value:\s+([^\n]+)')
    all = re.findall(reg, config_dump)
    # tidy emits the checker option with a "." prefix, but we need a ":"
    all = [(option[0].replace(".", ":"), option[1]) for option in all]
    return all


def parse_checker_config_new(config_dump):
    """
    Return the parsed clang-tidy config options as a list of
    (flag, default_value) tuples. This variant works
    for clang-tidy starting with version 15

    config_dump -- clang-tidy config options YAML dump in post-LLVM15 format.
    """
    try:
        data = yaml.safe_load(config_dump)
        if 'CheckOptions' not in data:
            return None

        return [[key.replace(".", ":"), value]
                for (key, value) in data['CheckOptions'].items()]
    except ImportError:
        return None


def parse_checker_config(config_dump):
    """
    Return the parsed clang-tidy config options as a list of
    (flag, default_value) tuples.
    config_dump -- clang-tidy config options YAML dump.
    """
    result = parse_checker_config_old(config_dump)
    if not result:
        result = parse_checker_config_new(config_dump)

    return result


def parse_analyzer_config(config_dump):
    """
    Return the parsed clang-tidy analyzer options as a list of
    (flag, default_value) tuples.
    config_dump -- clang-tidy config options YAML dump.
    """
    return re.findall(r'^(\S+):\s+(\S+)$', config_dump, re.MULTILINE)


def get_diagtool_bin():
    """
    Return full path of diagtool.

    Select clang binary, check for a 'diagtool' binary next to the selected
    clang binary and return full path of this binary if it exists.
    """
    context = analyzer_context.get_context()
    clang_bin = context.analyzer_binaries.get(ClangSA.ANALYZER_NAME)

    if not clang_bin:
        return None

    # Resolve symlink.
    clang_bin = os.path.realpath(clang_bin)

    # Find diagtool next to the clang binary.
    diagtool_bin = os.path.join(os.path.dirname(clang_bin), 'diagtool')
    if os.path.exists(diagtool_bin):
        return diagtool_bin

    LOG.debug("'diagtool' can not be found next to the clang binary (%s)!",
              clang_bin)


def get_warnings(env=None):
    """
    Returns list of warning flags by using diagtool.
    """
    diagtool_bin = get_diagtool_bin()

    if not diagtool_bin:
        return []

    try:
        result = subprocess.check_output(
            [diagtool_bin, 'tree'],
            env=env,
            universal_newlines=True,
            encoding="utf-8",
            errors="ignore")
        return [w[2:] for w in result.split()
                if w.startswith("-W") and w != "-W"]
    except subprocess.CalledProcessError as exc:
        LOG.error("'diagtool' encountered an error while retrieving the "
                  "checker list. If you are using a custom compiled clang, "
                  "you may have forgotten to build the 'diagtool' target "
                  "alongside 'clang' and 'clang-tidy'! Error message: %s",
                  exc.output)

        raise
    except OSError:
        raise


def _add_asterisk_for_group(
    subset_checkers: Iterable[str],
    all_checkers: Set[str]
) -> List[str]:
    """
    Since CodeChecker interprets checker name prefixes as checker groups, they
    have to be added a '*' joker character when using them at clang-tidy
    -checks flag. This function adds a '*' for each item in "checkers" if it's
    a checker group, i.e. identified as a prefix for any checker name in
    "all_checkers".
    For example "readability-container" is a prefix of multiple checkers, so
    this is converted to "readability-container-*". On the other hand
    "performance-trivially-destructible" is a full checker name, so it remains
    as is.
    """
    def is_group_prefix_of(prefix: str, long: str) -> bool:
        """
        Returns True if a checker(-group) name is prefix of another
        checker name. For example bugprone-string is prefix of
        bugprone-string-constructor but not of
        bugprone-stringview-nullptr.
        """
        prefix_split = prefix.split('-')
        long_split = long.split('-')
        return prefix_split == long_split[:len(prefix_split)]

    def need_asterisk(checker: str) -> bool:
        return any(
            is_group_prefix_of(checker, long) and checker != long
            for long in all_checkers)

    result = []

    for checker in subset_checkers:
        result.append(checker + ('*' if need_asterisk(checker) else ''))

    return result


def parse_version(tidy_output):
    """
    Parse clang-tidy version output and return the version number.
    """
    version_re = re.compile(r'.*version (?P<version>[\d\.]+)', re.S)
    match = version_re.match(tidy_output)
    if match:
        return match.group('version')


class ClangTidy(analyzer_base.SourceAnalyzer):
    """
    Constructs the clang tidy analyzer commands.
    """

    ANALYZER_NAME = 'clang-tidy'

    # Cache object for get_analyzer_checkers().
    __analyzer_checkers = None

    @classmethod
    def analyzer_binary(cls):
        return analyzer_context.get_context() \
            .analyzer_binaries[cls.ANALYZER_NAME]

    @classmethod
    def get_binary_version(self, environ, details=False) -> str:
        # No need to LOG here, we will emit a warning later anyway.
        if not self.analyzer_binary():
            return None

        version = [self.analyzer_binary(), '--version']
        try:
            output = subprocess.check_output(version,
                                             env=environ,
                                             universal_newlines=True,
                                             encoding="utf-8",
                                             errors="ignore")
            if details:
                return output.strip()
            return parse_version(output)
        except (subprocess.CalledProcessError, OSError) as oerr:
            LOG.warning("Failed to get analyzer version: %s",
                        ' '.join(version))
            LOG.warning(oerr)

        return None

    def add_checker_config(self, checker_cfg):
        LOG.error("Not implemented yet")

    @classmethod
    def get_analyzer_checkers(cls):
        """
        Return the list of the all of the supported checkers.
        """
        try:
            if cls.__analyzer_checkers:
                return cls.__analyzer_checkers

            environ = analyzer_context.get_context().analyzer_env
            result = subprocess.check_output(
                [cls.analyzer_binary(), "-list-checks", "-checks=*"],
                env=environ,
                universal_newlines=True,
                encoding="utf-8",
                errors="ignore")
            checker_description = parse_checkers(result)

            checker_description.extend(
                ("clang-diagnostic-" + warning, "")
                for warning in get_warnings(environ))

            cls.__analyzer_checkers = checker_description

            return checker_description
        except (subprocess.CalledProcessError, OSError):
            return []

    @classmethod
    def get_checker_config(cls):
        """
        Return the checker configuration of the all of the supported checkers.
        """
        try:
            result = subprocess.check_output(
                [cls.analyzer_binary(), "-dump-config", "-checks=*"],
                env=analyzer_context.get_context().analyzer_env,
                universal_newlines=True,
                encoding="utf-8",
                errors="ignore")
            return parse_checker_config(result)
        except (subprocess.CalledProcessError, OSError):
            return []

    @classmethod
    def get_analyzer_config(cls):
        """
        Return the analyzer configuration with all checkers enabled.
        """
        try:
            result = subprocess.check_output(
                [cls.analyzer_binary(), "-dump-config", "-checks=*"],
                env=analyzer_context.get_context().analyzer_env,
                universal_newlines=True,
                encoding="utf-8",
                errors="ignore")
            return parse_analyzer_config(result)
        except (subprocess.CalledProcessError, OSError):
            return []

    def get_checker_list(self, config) -> Tuple[List[str], List[str]]:
        """
        Return a list of checkers and warnings what needs to be enabled during
        analysis.

        If 'Checks' option is specified through '--analyzer-config'
        the return value will be a tuple of empty lists which means do not turn
        checkers explicitly. "clang-analyzer-*" is an exception, because we
        want to disable these even if analyzer config is given. If we wouldn't
        disable this group then some ClangSA checkers would be invoked by
        clang-tidy and in some cases that would cause analysis error due to
        some ClangSA bug.
        """
        # clang-tidy emits reports from its check in the same format as
        # compiler diagnostics (like unused variables, etc). This makes it a
        # little difficult to distinguish compiler warnings and clang-tidy
        # check warnings. The only clue is that compiler warnings are emitted
        # as if they came from a check called clang-diagnostic- (e.g.
        # -Wunused-variable will emit a warning under the name
        # clang-diagnostic-unused-variable).

        # There are two ways to disable a compiler warning in clang-tidy,
        # either by -Wno- or -checks=-clang-diagnostic- (note the dash before
        # clang-diagnostic!). However, there is only one way to enable them:
        # through -W. Using -checks=clang-diagnostic- does not enable the
        # warning, but undoes -checks=-clang-diagnostic-.

        # Since we disable all checks by default via -checks=-*, in order to
        # enable a compiler warning, we first have to undo the -checks level
        # disable and then enable it, so we need both
        # -checks=compiler-diagnostic- and -W.
        compiler_warnings = list()
        enabled_checkers = list()

        has_checker_config = \
            config.checker_config and config.checker_config != '{}'

        # Config handler stores which checkers are enabled or disabled.
        for checker_name, value in config.checks().items():
            state, _ = value

            warning_name, warning_type = \
                get_compiler_warning_name_and_type(checker_name)

            # This warning must be given a parameter separated by either '=' or
            # space. This warning is not supported as a checker name so its
            # special usage is avoided.
            if warning_name and warning_name.startswith('frame-larger-than'):
                continue

            if warning_name is not None:
                # -W and clang-diagnostic- are added as compiler warnings.
                if warning_type == CheckerType.compiler:
                    LOG.warning("As of CodeChecker v6.22, the following usage"
                                f"of '{checker_name}' compiler warning as a "
                                "checker name is deprecated, please use "
                                f"'clang-diagnostic-{checker_name[1:]}' "
                                "instead.")
                    if state == CheckerState.enabled:
                        compiler_warnings.append('-W' + warning_name)
                        enabled_checkers.append(checker_name)
                    elif state == CheckerState.disabled:
                        if config.enable_all:
                            LOG.warning("Disabling compiler warning with "
                                        f"compiler flag '-d W{warning_name}' "
                                        "is not supported.")
                # If a clang-diagnostic-... is enabled add it as a compiler
                # warning as -W..., if it is disabled, tidy can suppress when
                # specified in the -checks parameter list, so we add it there
                # as -clang-diagnostic-... .
                elif warning_type == CheckerType.analyzer:
                    if state == CheckerState.enabled:
                        compiler_warnings.append('-W' + warning_name)
                        enabled_checkers.append(checker_name)
                    else:
                        compiler_warnings.append('-Wno-' + warning_name)

                continue

            if state == CheckerState.enabled:
                enabled_checkers.append(checker_name)

        # By default all checkers are disabled and the enabled ones are added
        # explicitly.
        checkers = ['-*']

        checkers += _add_asterisk_for_group(
            enabled_checkers,
            set(x[0] for x in ClangTidy.get_analyzer_checkers()))

        # -checks=-clang-analyzer-* option is added to the analyzer command by
        # default except when all analyzer config options come from .clang-tidy
        # file. The content of this file overrides every other custom config
        # that is specific to clang-tidy. Compiler warnings however are flags
        # for the compiler, clang-tidy is just capable to emit them in its own
        # format.
        if config.analyzer_config.get('take-config-from-directory') == 'true':
            return [], compiler_warnings

        if has_checker_config:
            try:
                # Is is possible that the value of config.checker_config
                # is not a valid JSON string because the keys/values are
                # quoted with single quotes instead of double quotes. For
                # this reason we can't use the json.loads function to parse
                # this string but we need to use 'literal_eval' function to
                # safely evaluate the expression which will be a
                # valid dictionary.
                checker_cfg = ast.literal_eval(config.checker_config.strip())
                if checker_cfg.get('Checks'):
                    return [], compiler_warnings
            except SyntaxError as ex:
                LOG.debug("Invalid checker configuration: %s. Error: %s",
                          config.checker_config, ex)

        return checkers, compiler_warnings

    def construct_analyzer_cmd(self, result_handler):
        """ Contruct command which will be executed on analysis. """
        try:
            config = self.config_handler

            analyzer_cmd = [ClangTidy.analyzer_binary()]

            checks, compiler_warnings = self.get_checker_list(config)

            if checks:
                # The invocation should end in a Popen call with shell=False,
                # so no globbing should occur even if the checks argument
                # contains characters that would trigger globbing in the shell.
                analyzer_cmd.append("-checks=%s" % ','.join(checks))

            analyzer_cmd.extend(config.analyzer_extra_arguments)

            if config.checker_config and config.checker_config != '{}':
                analyzer_cmd.append("-config=" + config.checker_config)

            analyzer_cmd.append(self.source_file)

            analyzer_cmd.extend(['--export-fixes', result_handler.fixit_file])

            analyzer_cmd.append("--")

            analyzer_cmd.append('-Qunused-arguments')

            # Enable these compiler warnings by default.
            if not config.enable_all:
                analyzer_cmd.extend(['-Wno-everything'])

            compile_lang = self.buildaction.lang

            if not has_flag('-x', analyzer_cmd):
                analyzer_cmd.extend(['-x', compile_lang])

            if not has_flag('--target', analyzer_cmd) and \
                    self.buildaction.target != "":
                analyzer_cmd.append(f"--target={self.buildaction.target}")

            if not has_flag('-arch', analyzer_cmd) and \
                    self.buildaction.arch != "":
                analyzer_cmd.extend(["-arch", self.buildaction.arch])

            analyzer_cmd.extend(self.buildaction.analyzer_options)

            analyzer_cmd.extend(prepend_all(
                '-isystem',
                self.buildaction.compiler_includes))

            if not has_flag('-std', analyzer_cmd) and not \
                    has_flag('--std', analyzer_cmd):
                analyzer_cmd.append(self.buildaction.compiler_standard)

            if config.enable_all:
                analyzer_cmd.append("-Weverything")
                analyzer_cmd.extend(
                    filter(lambda x: x.startswith('-Wno-'), compiler_warnings))
            else:
                analyzer_cmd.extend(compiler_warnings)

            return analyzer_cmd

        except Exception as ex:
            LOG.error(ex)
            return []

    def get_analyzer_mentioned_files(self, output):
        """
        Parse Clang-Tidy's output to generate a list of files that were
        mentioned in the standard output or standard error.
        """

        if not output:
            return set()

        # A line mentioning a file in Clang-Tidy's output looks like this:
        # /home/.../.cpp:L:C: warning: foobar.
        regex = re.compile(
            # File path followed by a ':'.
            r'^(?P<path>[\S ]+?):'
            # Line number followed by a ':'.
            r'(?P<line>\d+?):'
            # Column number followed by a ':' and a space.
            r'(?P<column>\d+?): ')

        paths = []

        for line in output.splitlines():
            match = re.match(regex, line)
            if match:
                paths.append(match.group('path'))

        return set(paths)

    @classmethod
    def resolve_missing_binary(cls, configured_binary, environ):
        """
        In case of the configured binary for the analyzer is not found in the
        PATH, this method is used to find a callable binary.
        """

        LOG.debug("%s not found in path for ClangTidy!", configured_binary)

        if os.path.isabs(configured_binary):
            # Do not autoresolve if the path is an absolute path as there
            # is nothing we could auto-resolve that way.
            return False

        # clang-tidy, clang-tidy-5.0, ...
        clangtidy = env.get_binary_in_path(['clang-tidy'],
                                           r'^clang-tidy(-\d+(\.\d+){0,2})?$',
                                           environ)

        if clangtidy:
            LOG.debug("Using '%s' for Clang-tidy!", clangtidy)
        return clangtidy

    @classmethod
    def is_binary_version_incompatible(cls, environ):
        """
        We support pretty much every Clang-Tidy version.
        """
        return None

    def construct_result_handler(self, buildaction, report_output,
                                 skiplist_handler):
        """
        See base class for docs.
        """
        report_hash = self.config_handler.report_hash
        res_handler = result_handler.ClangTidyResultHandler(
            buildaction, report_output, report_hash)

        res_handler.skiplist_handler = skiplist_handler
        return res_handler

    @classmethod
    def construct_config_handler(cls, args):
        handler = config_handler.ClangTidyConfigHandler()
        handler.report_hash = args.report_hash \
            if 'report_hash' in args else None

        # FIXME We cannot get the resource dir from the clang-tidy binary,
        # therefore we get a sibling clang binary which of clang-tidy.
        # TODO Support "clang-tidy -print-resource-dir" .
        try:
            with open(args.tidy_args_cfg_file, 'r', encoding='utf-8',
                      errors='ignore') as tidy_cfg:
                handler.analyzer_extra_arguments = \
                    re.sub(r'\$\((.*?)\)',
                           env.replace_env_var(args.tidy_args_cfg_file),
                           tidy_cfg.read().strip())
                handler.analyzer_extra_arguments = \
                    shlex.split(handler.analyzer_extra_arguments)
        except IOError as ioerr:
            LOG.debug_analyzer(ioerr)
        except AttributeError as aerr:
            # No clang tidy arguments file was given in the command line.
            LOG.debug_analyzer(aerr)

        analyzer_config = {}
        # TODO: This extra "isinstance" check is needed for
        # CodeChecker analyzers --analyzer-config. This command also
        # runs this function in order to construct a config handler.
        if 'analyzer_config' in args and \
                isinstance(args.analyzer_config, list):
            for cfg in args.analyzer_config:
                if cfg.analyzer == cls.ANALYZER_NAME:
                    analyzer_config[cfg.option] = cfg.value

        # Reports in headers are hidden by default in clang-tidy. Re-enable it
        # if the analyzer doesn't provide any other config options.
        if not analyzer_config:
            analyzer_config["HeaderFilterRegex"] = ".*"

        # If both --analyzer-config and -config (in --tidyargs) is given then
        # these need to be merged. Since "HeaderFilterRegex" has a default
        # value in --analyzer-config, we take --tidyargs stronger so user can
        # overwrite its value.
        for i, extra_arg in enumerate(handler.analyzer_extra_arguments):
            if not extra_arg.startswith('-config'):
                continue

            # -config flag can be together or separate from its argument:
            # "-config blabla" vs. "-config=blabla"
            if extra_arg == '-config':
                arg = handler.analyzer_extra_arguments[i + 1]
                arg_num = 2
            else:
                arg = extra_arg[len('-config='):]
                arg_num = 1

            analyzer_config.update(json.loads(arg))
            del handler.analyzer_extra_arguments[i:i + arg_num]
            break

        # TODO: This extra "isinstance" check is needed for
        # CodeChecker checkers --checker-config. This command also
        # runs this function in order to construct a config handler.
        if 'checker_config' in args and isinstance(args.checker_config, list):
            check_options = []

            for cfg in args.checker_config:
                if cfg.analyzer == cls.ANALYZER_NAME:
                    check_options.append({
                        'key': f"{cfg.checker}.{cfg.option}",
                        'value': cfg.value})
            analyzer_config['CheckOptions'] = check_options
        else:
            try:
                with open(args.tidy_config, 'r',
                          encoding='utf-8', errors='ignore') as tidy_config:
                    handler.checker_config = tidy_config.read()
            except IOError as ioerr:
                LOG.debug_analyzer(ioerr)
            except AttributeError as aerr:
                # No clang tidy config file was given in the command line.
                LOG.debug_analyzer(aerr)

        handler.analyzer_config = analyzer_config

        # 'take-config-from-directory' is a special option which let the user
        # to use the '.clang-tidy' config files. It will disable analyzer and
        # checker configuration options.
        if not handler.checker_config and \
                analyzer_config.get('take-config-from-directory') != 'true':
            handler.checker_config = json.dumps(analyzer_config)

        checkers = ClangTidy.get_analyzer_checkers()

        try:
            cmdline_checkers = args.ordered_checkers
        except AttributeError:
            LOG.debug_analyzer('No checkers were defined in '
                               'the command line for %s',
                               cls.ANALYZER_NAME)
            cmdline_checkers = []

        handler.initialize_checkers(
            checkers,
            cmdline_checkers,
            'enable_all' in args and args.enable_all)

        return handler



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
""" Suppress file format.

# This is the old format.
123324353456463442341242342343#1 || bug hash comment

# This is the new format.
123324353456463442341242342343#1 || filename || bug hash comment

After removing the hash_value_type the generated format is:
123324353456463442341242342343 || filename || bug hash comment

For backward compatibility the hash_value_type is an optional filed.
"""


import codecs
import os
import re

from codechecker_common.logger import get_logger
from codechecker_common.source_code_comment_handler import \
    SourceCodeCommentHandler

LOG = get_logger('system')

COMMENT_SEPARATOR = '||'
HASH_TYPE_SEPARATOR = '#'


def get_suppress_data(suppress_file):
    """
    Process a file object for suppress information.
    """

    old_format_pattern = r"^(?P<bug_hash>[\d\w]{32})" \
        r"(\#(?P<bug_hash_type>\d))?\s*\|\|\s*(?P<comment>[^\|]*)$"
    old_format = re.compile(old_format_pattern, re.UNICODE)

    new_format_pattern = r"^(?P<bug_hash>[\d\w]{32})" \
        r"(\#(?P<bug_hash_type>\d))?\s*\|\|\s*" \
        r"(?P<file_name>[^\\\|]+)\s*\|\|\s*(?P<comment>[^\|]*)$"
    new_format = re.compile(new_format_pattern, re.UNICODE)

    markers = '|'.join(SourceCodeCommentHandler.source_code_comment_markers)
    src_suppress_format_pattern = r"^(?P<bug_hash>[\d\w]{32})" \
        r"(\#(?P<bug_hash_type>\d))?\s*\|\|\s*" \
        r"(?P<file_name>[^\\\|]+)\s*\|\|\s*(?P<comment>[^\|]*)\s*\|\|\s*" \
        r"(?P<status>[" + markers + "]*)$"
    src_suppress_format = re.compile(src_suppress_format_pattern, re.UNICODE)

    suppress_data = []

    for line in suppress_file:

        src_suppress_format_match = re.match(src_suppress_format, line.strip())
        if src_suppress_format_match:
            LOG.debug('Match for source code suppress entry format:')
            src_suppress_format_match = src_suppress_format_match.groupdict()
            LOG.debug(src_suppress_format_match)
            suppress_data.append((src_suppress_format_match['bug_hash'],
                                  src_suppress_format_match['file_name'],
                                  src_suppress_format_match['comment'],
                                  src_suppress_format_match['status']))
            continue

        new_format_match = re.match(new_format, line.strip())
        if new_format_match:
            LOG.debug('Match for new suppress entry format:')
            new_format_match = new_format_match.groupdict()
            LOG.debug(new_format_match)
            suppress_data.append((new_format_match['bug_hash'],
                                  new_format_match['file_name'],
                                  new_format_match['comment'],
                                  'false_positive'))
            continue

        old_format_match = re.match(old_format, line.strip())
        if old_format_match:
            LOG.debug('Match for old suppress entry format:')
            old_format_match = old_format_match.groupdict()
            LOG.debug(old_format_match)
            suppress_data.append((old_format_match['bug_hash'],
                                  '',  # empty file name
                                  old_format_match['comment'],
                                  'false_positive'))
            continue

        if line.strip() != '':
            LOG.warning('Malformed suppress line: %s', line)

    return suppress_data


def write_to_suppress_file(suppress_file, value, file_name, comment='',
                           status='false_positive'):
    comment = comment.decode('utf-8')

    LOG.debug('Processing suppress file: %s', suppress_file)

    try:
        with codecs.open(suppress_file, 'r', 'utf-8') as s_file:
            suppress_data = get_suppress_data(s_file)

        if not os.stat(suppress_file)[6] == 0:
            # File is not empty.

            res = [
                x for x in suppress_data if (
                    x[0] == value and x[1] == file_name) or (
                    x[0] == value and x[1] == '')]

            if res:
                LOG.debug("Already found in\n %s", suppress_file)
                return True

        s_file = codecs.open(suppress_file, 'a', 'utf-8')

        s_file.write(value + COMMENT_SEPARATOR +
                     file_name + COMMENT_SEPARATOR +
                     comment + COMMENT_SEPARATOR +
                     status + '\n')
        s_file.close()

        return True

    except Exception as ex:
        LOG.error(str(ex))
        LOG.error("Failed to write: %s", suppress_file)
        return False



#!/usr/bin/env python3
# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import argparse
import os
import sys

# If we run this script in an environment where 'codechecker_report_converter'
# module is not available we should add the grandparent directory of this file
# to the system path.
if __name__ == '__main__':
    sys.path.insert(0, os.path.abspath(
        os.path.join(__file__, *[os.path.pardir] * 4)))

from codechecker_report_converter.report.output.html.html import HtmlBuilder, \
    parse


def __add_arguments_to_parser(parser: argparse.ArgumentParser):
    parser.add_argument('input',
                        type=str,
                        nargs='+',
                        metavar='file/folder',
                        help="Analyzer result file(s) or folders containing "
                             "analysis results which should be parsed.")

    parser.add_argument('-o', '--output',
                        dest="output_dir",
                        required=True,
                        help="Generate HTML output files in the given folder.")

    curr_file_dir = os.path.dirname(os.path.realpath(__file__))
    parser.add_argument('-l', '--layout',
                        dest="layout_dir",
                        required=False,
                        default=os.path.join(curr_file_dir, 'static'),
                        help="Directory which contains dependency HTML, CSS "
                             "and JavaScript files.")


def main():
    """ Report to HTML main command line. """
    parser = argparse.ArgumentParser(
        prog="plist-to-html",
        description="Parse and create HTML files from one or more analyzer "
                    "result files.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    __add_arguments_to_parser(parser)

    args = parser.parse_args()

    if isinstance(args.input, str):
        args.input = [args.input]

    # Source files which modification time changed since the last analysis.
    changed_source_files = set()

    html_builder = HtmlBuilder(args.layout_dir)
    for input_path in args.input:
        changed_files = parse(input_path, args.output_dir, args.layout_dir,
                              html_builder)
        changed_source_files.update(changed_files)

    html_builder.create_index_html(args.output_dir)
    html_builder.create_statistics_html(args.output_dir)

    print('\nTo view statistics in a browser run:\n> firefox {0}'.format(
        os.path.join(args.output_dir, 'statistics.html')))

    print('\nTo view the results in a browser run:\n> firefox {0}'.format(
        os.path.join(args.output_dir, 'index.html')))

    if changed_source_files:
        changed_files = '\n'.join([' - ' + f for f in changed_source_files])
        print("\nThe following source file contents changed since the "
              "latest analysis:\n{0}\nPlease analyze your project again to "
              "update the reports!".format(changed_files))


if __name__ == "__main__":
    main()



"""
Source components

Revision ID: 3793e361a752
Revises:     080349e895d7
Create Date: 2018-04-20 09:29:24.072720
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = '3793e361a752'
down_revision = '080349e895d7'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'source_components',
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('value', sa.Binary(), nullable=False),
        sa.Column('description', sa.Text(), nullable=True),
        sa.Column('username', sa.String(), nullable=True),
        sa.PrimaryKeyConstraint('name', name=op.f('pk_source_components'))
    )


def downgrade():
    op.drop_table('source_components')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
import os
import sys

from alembic import context
from sqlalchemy import engine_from_config, pool

from codechecker_server.database.run_db_model import Base
from codechecker_server.migrations.logging import set_logger_database_name

# This is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Add model's MetaData object here for 'autogenerate' support.
sys.path.insert(0, os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", "..", "..")))

target_metadata = Base.metadata

schema = "report"

# Other values from the config, defined by the needs of env.py, can be
# acquired: my_important_option = config.get_main_option("my_important_option")


def run_migrations_online():
    """
    Run migrations in 'online' mode.

    In this scenario we need to create an Engine and associate a connection
    with the context.
    """
    connectable = config.attributes.get('connection', None)
    if connectable is None:
        connectable = engine_from_config(
            config.get_section(config.config_ini_section),
            prefix='sqlalchemy.',
            poolclass=pool.NullPool)

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata
        )
        set_logger_database_name(schema,
                                 config.attributes.get("database_name"))

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    raise NotImplementedError(f"Offline '{schema}' migration is not possible!")
run_migrations_online()



"""
Store information about enabled and disabled checkers for a run.

Revision ID: c3dad71f8e6b
Revises:     9d956a0fae8d
Create Date: 2023-10-20 14:11:48.371981
"""
from logging import getLogger
from typing import Dict, Tuple

from alembic import op
import sqlalchemy as sa
from sqlalchemy.orm import Session
from sqlalchemy.sql.expression import and_, func, not_
from sqlalchemy.ext.automap import automap_base

from codechecker_report_converter.report import FakeChecker, UnknownChecker


# Revision identifiers, used by Alembic.
revision = 'c3dad71f8e6b'
down_revision = '9d956a0fae8d'
branch_labels = None
depends_on = None

REPORT_UPDATE_CHUNK_SIZE = 1_000_000


def upgrade():
    # Note: The instantiation of the LOG variable *MUST* stay here so that it
    # uses the facilities that are sourced from the Alembic env.py.
    # Symbols created on the module-level are created *before* Alembic's env.py
    # had loaded.
    LOG = getLogger("migration/report")
    dialect = op.get_context().dialect.name
    conn = op.get_bind()

    def normalise_report_analyzer_and_checker_names():
        # Unknown analyzers and checkers might be represented in the database
        # with several values all meaning the same, gathered over the years.
        # Notably, the DEFAULT value for 'analyzer_name' is "unknown" prior
        # to the upgrade. While 'checker_id' can be NULL, it could be an empty
        # string, or a "NOT FOUND" created by the report-converter, for reports
        # newer than the existence of the report-converter.
        # These values are normalised such that in the following, when the
        # foreign key-based look-up is added to the schema, their new
        # 'checker_id' will all point to the single "UNKNOWN/NOT FOUND" case.
        analyzer_name_affected = conn.execute(f"""
            UPDATE reports
            SET analyzer_name = '{UnknownChecker[0]}'
            WHERE analyzer_name = ''
                OR LOWER(analyzer_name) = 'unknown'
            ;
        """).rowcount
        if analyzer_name_affected:
            LOG.info("Normalising 'reports'... %d unknown 'analyzer_name'.",
                     analyzer_name_affected)

        checker_id_affected = conn.execute(f"""
            UPDATE reports
            SET checker_id = '{UnknownChecker[1]}'
            WHERE checker_id IS NULL
                OR checker_id = ''
                OR LOWER(checker_id) = 'not found'
                OR LOWER(checker_id) = 'unknown'
            ;
        """).rowcount
        if checker_id_affected:
            LOG.info("Normalising 'reports'... %d unknown 'checker_id'.",
                     checker_id_affected)

        if dialect == "postgresql":
            # The changes of the above queries must be committed prematurely,
            # as later schema changes (such as renaming and dropping the
            # columns normalised now) would otherwise fail with an error:
            #
            # > sqlalchemy.exc.OperationalError: (psycopg2.errors.ObjectInUse)
            # >    cannot ALTER TABLE "reports" because it has pending
            # >    trigger events.
            # > [SQL: ALTER TABLE reports ADD COLUMN checker_id INTEGER
            # >       DEFAULT '0' NOT NULL]
            # > [SQL: ALTER TABLE reports DROP COLUMN checker_name]
            #
            # There is no good sequencing of these changes that would resolve
            # this while keeping the entire migration in one transaction.
            # However, these changes are not destructive because no data is
            # lost, only the representation slightly changed.
            conn.execute("COMMIT;")
            conn.execute("START TRANSACTION;")

    def create_new_tables():
        op.create_table(
            "checkers",
            sa.Column("id", sa.Integer(), autoincrement=True, nullable=False),
            sa.Column("analyzer_name", sa.String(), nullable=True),
            sa.Column("checker_name", sa.String(), nullable=True),
            sa.Column("severity", sa.Integer()),
            sa.PrimaryKeyConstraint("id", name=op.f("pk_checkers")),
            sa.UniqueConstraint("analyzer_name", "checker_name",
                                name=op.f("uq_checkers_analyzer_name"))
        )
        op.create_index(
            op.f("ix_checkers_severity"),
            "checkers",
            ["severity"],
            unique=False
        )

        fk_analysisinfo_checkers_name = \
            "fk_analysis_info_checkers_analysis_info_id_analysis_info"
        op.create_table(
            "analysis_info_checkers",
            sa.Column("analysis_info_id", sa.Integer(), nullable=False),
            sa.Column("checker_id", sa.Integer(), nullable=False),
            sa.Column("enabled", sa.Boolean(), nullable=True),
            sa.ForeignKeyConstraint(
                ["analysis_info_id"], ["analysis_info.id"],
                name=op.f(fk_analysisinfo_checkers_name),
                ondelete="CASCADE", initially="DEFERRED",
                deferrable=True),
            sa.ForeignKeyConstraint(
                ["checker_id"], ["checkers.id"],
                name=op.f("fk_analysis_info_checkers_checker_id_checkers"),
                ondelete="RESTRICT", initially="DEFERRED", deferrable=True),
            sa.PrimaryKeyConstraint("analysis_info_id", "checker_id",
                                    name=op.f("pk_analysis_info_checkers"))
        )

    def add_new_checker_id_column():
        # Upgrade the 'reports' table to use the 'checkers' foreign look-up
        # instead of containing the strings allocated locally with the record.
        col_reports_checker_id = sa.Column("checker_id", sa.Integer(),
                                           nullable=False, server_default="1")

        if dialect == "sqlite":
            with op.batch_alter_table("reports", recreate="never") as ba:
                ba.alter_column("checker_id", new_column_name="checker_name")
            # Due to the rename reusing the previous name, these can't merge.
            with op.batch_alter_table("reports", recreate="never") as ba:
                ba.add_column(col_reports_checker_id)
        else:
            op.alter_column("reports", "checker_id",
                            new_column_name="checker_name")
            op.add_column("reports", col_reports_checker_id)

    def add_checkers():
        # Pre-allocate IDs in the look-up table for all checkers that were
        # encountered according to the currently present reports in the DB.
        Base = automap_base()
        Base.prepare(conn, reflect=True)
        Report = Base.classes.reports  # 'reports' is the table name!
        Checker = Base.classes.checkers

        db = Session(bind=conn)
        db.add(Checker(analyzer_name=FakeChecker[0],
                       checker_name=FakeChecker[1],
                       severity=0))
        db.add(Checker(analyzer_name=UnknownChecker[0],
                       checker_name=UnknownChecker[1],
                       severity=0))

        count = db.query(Report.id).count()
        if not count:
            return

        # It is possible that the same (analyzer_name, checker_name) query,
        # when GROUPed BY, will still have multiple reports with distinct
        # severities. Adding all entries here would result in the UNIQUE
        # constraint violation. Luckily, severities are numeric, and thus can
        # be sorted, and saying that for each checkers the largest severity
        # should be the new severity is a reasonable assumption.
        LOG.info("Preparing to fill 'checkers' from %d 'reports'...", count)
        checker_count = 0
        for chk in db.query(Report.analyzer_name,
                            Report.checker_name,
                            func.max(Report.severity).label("severity")) \
                .filter(not_(and_(
                    Report.analyzer_name == UnknownChecker[0],
                    Report.checker_name == UnknownChecker[1]))) \
                .group_by(Report.analyzer_name,
                          Report.checker_name) \
                .all():
            db.add(Checker(analyzer_name=chk.analyzer_name,
                           checker_name=chk.checker_name,
                           severity=chk.severity))
            checker_count += 1

        db.commit()
        LOG.info("Done filling 'checkers', %d unique entries.", checker_count)

    def upgrade_reports():
        Base = automap_base()
        Base.prepare(conn, reflect=True)
        Report = Base.classes.reports  # 'reports' is the table name!
        Checker = Base.classes.checkers

        db = Session(bind=conn)
        report_count = db.query(Report.id).count()
        if not report_count:
            return

        LOG.info("Upgrading %d 'reports' to refer 'checkers',"
                 " in batches of %d...",
                 report_count, REPORT_UPDATE_CHUNK_SIZE)
        num_batches = report_count // REPORT_UPDATE_CHUNK_SIZE + 1

        def _print_progress(batch: int):
            LOG.info("[%d/%d] Upgrading 'reports'... (%d-%d) %.0f%% done.",
                     batch, num_batches,
                     (REPORT_UPDATE_CHUNK_SIZE * i) + 1,
                     (REPORT_UPDATE_CHUNK_SIZE * (i + 1))
                     if batch < num_batches else report_count,
                     float(batch) / num_batches * 100)

        fake_chk_id = db.query(Checker.id) \
            .filter(Checker.analyzer_name == FakeChecker[0],
                    Checker.checker_name == FakeChecker[1]) \
            .scalar()
        for i in range(0, num_batches):
            # FIXME: "UPDATE ... SET ... FROM ..." is only supported starting
            # with SQLite version 3.33.0 (2020-08-14). Until this version
            # reaches LTS maturity (Ubuntu 20.04 LTS comes with 3.31.0, raising
            # a syntax error on the "FROM" in the "UPDATE" query), this
            # branching here needs to stay.
            if dialect == "sqlite":
                db.execute(f"""
                    UPDATE reports
                    SET
                      checker_id = (
                        SELECT checkers.id
                        FROM checkers
                        WHERE checkers.analyzer_name = reports.analyzer_name
                          AND checkers.checker_name = reports.checker_name
                      )
                    WHERE reports.id IN (
                        SELECT reports.id
                        FROM reports
                        WHERE reports.checker_id = {fake_chk_id}
                          AND reports.analyzer_name != '{FakeChecker[0]}'
                          AND reports.checker_name != '{FakeChecker[1]}'
                        LIMIT {REPORT_UPDATE_CHUNK_SIZE}
                      )
                    ;
                """)
            else:
                db.execute(f"""
                    UPDATE reports
                    SET
                      checker_id = checkers.id
                    FROM checkers
                    WHERE checkers.analyzer_name = reports.analyzer_name
                      AND checkers.checker_name = reports.checker_name
                      AND reports.id IN (
                        SELECT reports.id
                        FROM reports
                        WHERE reports.checker_id = {fake_chk_id}
                          AND reports.analyzer_name != '{FakeChecker[0]}'
                          AND reports.checker_name != '{FakeChecker[1]}'
                        LIMIT {REPORT_UPDATE_CHUNK_SIZE}
                      )
                    ;
                """)
            _print_progress(i + 1)

        db.commit()
        LOG.info("Done upgrading 'reports'.")

    def drop_reports_table_columns():
        if dialect == "sqlite":
            op.execute("PRAGMA foreign_keys=OFF")

            with op.batch_alter_table("reports", recreate="never") as ba:
                # FIXME: Allowing recreate="auto" (the default) here would
                # result in SQLAlchemy creating a new 'reports' table, which
                # will break the database as the incoming constraints
                # FOREIGN KEYing against 'reports' will essentially TRUNCATE
                # due to the ON DELETE CASCADE markers, e.g., in
                # 'bug_path_events'. This manifests as the Python-based
                # database cleanup routine essentially wiping all reports off
                # the database, following this migration.
                #
                # Until SQLite 3.35.0 is widely available (Ubuntu 22.04, see
                # also dabc6998b8f0_analysis_info_table.py), we side-step the
                # current problem with a simple rename. However, in general,
                # this is always bound to get bad, and a real solution is only
                # expectable from a **significant**, ground-up redesign of both
                # the database and the migration logic, and our processes.
                # (E.g., if we always assumed migrations can trash the database
                # because there are always back-ups, the whole problem with
                # "SQLite + transaction + pragma = no-op" would not be here.)
                #
                # It generally appears that the big issue here is the fact that
                # migration scripts execute as part of a TRANSACTION! As such,
                # the "PRAGMA foreign_keys=OFF" is useless. See the docs at
                #     http://sqlite.org/foreignkeys.html (quoted Jan 16, 2024)
                #
                # > It is not possible to enable or disable foreign key
                # > constraints in the middle of a multi-statement transaction
                # > (when SQLite is not in autocommit mode). Attempting to do
                # > so does not return an error; it simply has no effect.
                #
                # This means that the foreign keys are still enabled, and when
                # the migration commits, 'bug_path_events' and
                # 'bug_report_points' **WILL** TRUNCATE themselves as the batch
                # operation executed a "DROP TABLE reports;". These claims can
                # be re-verified later by executing the following steps:
                #   1. Change the database manually and set the incoming
                #      foreign keys to "ON DELETE RESTRICT" instead of
                #      "CASCADE".
                #   2. Run the migration. An exception will come from this
                #      context but will refer to an internal "DROP TABLE" stmt.
                #   3. Add an op.execute("COMMIT") before the PRAGMA call
                #      above.
                #   4. Re-run the migration and observe everything is good and
                #      there was no data loss whatsoever!

                # ba.drop_column("checker_name")
                ba.alter_column(
                    "checker_name",
                    new_column_name="checker_name_MOVED_TO_checkers")

                # These columns are deleted as this data is now available
                # through the 'checkers' lookup-table.
                # ba.drop_column("analyzer_name")
                # ba.drop_column("severity")
                ba.alter_column(
                    "analyzer_name",
                    new_column_name="analyzer_name_MOVED_TO_checkers")
                ba.alter_column(
                    "severity",
                    new_column_name="severity_MOVED_TO_checkers")

                # These columns are dropped because they rarely contained any
                # meaningful data with new informational value, and their
                # contents were never actually exposed on the API.
                # ba.drop_column("checker_cat")
                # ba.drop_column("bug_type")
                ba.alter_column("checker_cat",
                                new_column_name="checker_cat_UNUSED")
                ba.alter_column("bug_type",
                                new_column_name="bug_type_UNUSED")

            op.execute("PRAGMA foreign_keys=ON")
        else:
            op.drop_column("reports", "checker_name")
            op.drop_column("reports", "analyzer_name")
            op.drop_column("reports", "severity")
            op.drop_column("reports", "checker_cat")
            op.drop_column("reports", "bug_type")

    def upgrade_reports_table_constraints():
        ix_reports_checker_id = {
            "index_name": op.f("ix_reports_checker_id"),
            "columns": ["checker_id"],
            "unique": False
        }
        fk_reports_checker_id = {
            "constraint_name": op.f("fk_reports_checker_id_checkers"),
            "referent_table": "checkers",
            "local_cols": ["checker_id"],
            "remote_cols": ["id"],
            "deferrable": False,
            "ondelete": "RESTRICT"
        }
        if dialect == "sqlite":
            op.execute("PRAGMA foreign_keys=OFF")

            with op.batch_alter_table("reports", recreate="never") as ba:
                # Now that the values are filled, ensure that the constriants
                # are appropriately enforced.
                ba.create_index(**ix_reports_checker_id)
                # This should really be a FOREIGN KEY, but it is not possible
                # without recreating the entire 'reports' table, which breaks
                # other FOREIGN KEYs.
                # ba.create_foreign_key(**fk_reports_checker_id)

            op.execute("PRAGMA foreign_keys=ON")
        else:
            op.create_index(table_name="reports", **ix_reports_checker_id)
            op.create_foreign_key(source_table="reports",
                                  **fk_reports_checker_id)
            op.alter_column("reports", "checker_id", nullable=False,
                            server_default=None)

    normalise_report_analyzer_and_checker_names()
    create_new_tables()
    add_new_checker_id_column()
    add_checkers()
    upgrade_reports()
    drop_reports_table_columns()
    upgrade_reports_table_constraints()


def downgrade():
    LOG = getLogger("migration/report")
    dialect = op.get_context().dialect.name
    conn = op.get_bind()

    def restore_report_columns():
        col_reports_analyzer_name = sa.Column("analyzer_name",
                                              sa.String(), nullable=False,
                                              server_default="unknown")
        col_reports_checker_id = sa.Column("checker_id", sa.String())
        col_reports_checker_cat = sa.Column("checker_cat", sa.String())
        col_reports_bug_type = sa.Column("bug_type", sa.String())
        col_reports_severity = sa.Column("severity", sa.Integer())

        if dialect == "sqlite":
            op.execute("PRAGMA foreign_keys=OFF")

            with op.batch_alter_table("reports", recreate="never") as ba:
                # Recreation of FOREIGN KEYs is not possible without recreating
                # the entire 'reports' table.
                # ba.drop_constraint(op.f("fk_reports_checker_id_checkers"))
                ba.drop_index(op.f("ix_reports_checker_id"))
                ba.alter_column("checker_id",
                                new_column_name="checker_id_lookup")

            with op.batch_alter_table("reports", recreate="never") as ba:
                # Restore the columns that were deleted in this revision.
                # ba.add_column(col_reports_analyzer_name)
                # ba.add_column(col_reports_checker_id)
                # ba.add_column(col_reports_checker_cat)
                # ba.add_column(col_reports_bug_type)
                # ba.add_column(col_reports_severity)

                # FIXME: Until SQLite 3.35 is available and we can actually
                # delete columns without a table recreation, we can instead
                # just restore the existing partial data!
                ba.alter_column("analyzer_name_MOVED_TO_checkers",
                                new_column_name="analyzer_name")
                ba.alter_column("checker_name_MOVED_TO_checkers",
                                new_column_name="checker_id")
                ba.alter_column("severity_moved_to_checkers",
                                new_column_name="severity")
                ba.alter_column("checker_cat_UNUSED",
                                new_column_name="checker_cat")
                ba.alter_column("bug_type_UNUSED",
                                new_column_name="bug_type")

                op.execute("PRAGMA foreign_keys=ON")
        else:
            op.drop_constraint(op.f("fk_reports_checker_id_checkers"),
                               "reports")
            op.drop_index(op.f("ix_reports_checker_id"), "reports")
            op.alter_column("reports", "checker_id",
                            new_column_name="checker_id_lookup")

            op.add_column("reports", col_reports_analyzer_name)
            op.add_column("reports", col_reports_checker_id)
            op.add_column("reports", col_reports_checker_cat)
            op.add_column("reports", col_reports_bug_type)
            op.add_column("reports", col_reports_severity)

        LOG.info("Restored type of columns 'reports.bug_type', "
                 "'reports.checker_cat'. However, their contents can NOT "
                 "be restored to the original values, as those were "
                 "irrevocably lost during a previous schema upgrade. "
                 "Note, that these columns NEVER contained any actual "
                 "value that was accessible by users of the API, so "
                 "this is a technical note.")

    def downgrade_reports():
        Base = automap_base()
        Base.prepare(conn, reflect=True)
        Report = Base.classes.reports  # 'reports' is the table name!
        Checker = Base.classes.checkers

        db = Session(bind=conn)
        report_count = db.query(Report.id).count()
        if not report_count:
            return

        LOG.info("Downgrading %d 'reports' from 'checkers',"
                 " in batches of %d...",
                 report_count, REPORT_UPDATE_CHUNK_SIZE)
        num_batches = report_count // REPORT_UPDATE_CHUNK_SIZE + 1

        def _print_progress(batch: int):
            LOG.info("[%d/%d] Downgrading 'reports'... (%d-%d) %.0f%% done.",
                     batch, num_batches,
                     (REPORT_UPDATE_CHUNK_SIZE * i) + 1,
                     (REPORT_UPDATE_CHUNK_SIZE * (i + 1))
                     if batch < num_batches else report_count,
                     float(batch) / num_batches * 100)

        fake_chk_id = db.query(Checker.id) \
            .filter(Checker.analyzer_name == FakeChecker[0],
                    Checker.checker_name == FakeChecker[1]) \
            .scalar()
        for i in range(0, num_batches):
            # FIXME: "UPDATE ... SET ... FROM ..." is only supported starting
            # with SQLite version 3.33.0 (2020-08-14). Until this version
            # reaches LTS maturity (Ubuntu 20.04 LTS comes with 3.31.0, raising
            # a syntax error on the "FROM" in the "UPDATE" query), this
            # branching here needs to stay.
            if dialect == "sqlite":
                db.execute(f"""
                    UPDATE reports
                    SET
                      (analyzer_name, checker_id, severity,
                        checker_id_lookup) =
                        (SELECT analyzer_name, checker_name, severity,
                                '{fake_chk_id}'
                          FROM checkers
                          WHERE checkers.id = reports.checker_id_lookup)
                    WHERE reports.id IN (
                      SELECT reports.id
                      FROM reports
                      WHERE reports.checker_id_lookup != {fake_chk_id}
                      LIMIT {REPORT_UPDATE_CHUNK_SIZE}
                        )
                    ;
                """)
            else:
                db.execute(f"""
                    UPDATE reports
                    SET
                        analyzer_name = checkers.analyzer_name,
                        checker_id = checkers.checker_name,
                        severity = checkers.severity,
                        checker_id_lookup = {fake_chk_id}
                    FROM checkers
                    WHERE checkers.id = reports.checker_id_lookup
                        AND reports.id IN (
                            SELECT reports.id
                            FROM reports
                            WHERE reports.checker_id_lookup != {fake_chk_id}
                            LIMIT {REPORT_UPDATE_CHUNK_SIZE}
                        )
                    ;
                """)
            _print_progress(i + 1)

        db.commit()
        LOG.info("Done downgrading 'reports'.")

    def drop_checker_id_numeric_column():
        if dialect == "sqlite":
            with op.batch_alter_table("reports", recreate="never") as ba:
                # FIXME: SQLite >= 3.35 will allow DROP COLUMN...
                # ba.drop_column("checker_id_lookup")
                ba.alter_column("checker_id_lookup",
                                new_column_name="checker_id_lookup_UNUSED")
        else:
            op.drop_column("reports", "checker_id_lookup")

    def drop_new_tables():
        # Drop all tables and columns that were created in this revision.
        # This data is not needed anymore.
        op.drop_index(op.f("ix_checkers_severity"), "checkers")
        op.drop_table("analysis_info_checkers")
        op.drop_table("checkers")

    restore_report_columns()
    downgrade_reports()
    drop_checker_id_numeric_column()
    drop_new_tables()



"""
Add confidentiality

Revision ID: cf025b6d7998
Revises:     4db450cf38af
Create Date: 2021-09-08 13:07:08.891285
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = 'cf025b6d7998'
down_revision = '4db450cf38af'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('products',
                  sa.Column('confidentiality', sa.String(), nullable=True))


def downgrade():
    op.drop_column('products', 'confidentiality')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Check static analyzer and features on the host machine.
"""


import errno
import re
import subprocess
import tempfile

from codechecker_analyzer import analyzer_context
from codechecker_common.logger import get_logger

LOG = get_logger('analyzer')


def check_analyzer(compiler_bin, env):
    """
    Simple check if clang is available.
    """
    clang_version_cmd = [compiler_bin, '--version']
    LOG.debug_analyzer(' '.join(clang_version_cmd))
    try:
        res = subprocess.call(
            clang_version_cmd,
            env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf-8",
            errors="ignore")
        if not res:
            return True

        LOG.debug_analyzer('Failed to run: "%s"', ' '.join(clang_version_cmd))
        return False

    except OSError as oerr:
        if oerr.errno == errno.ENOENT:
            LOG.error(oerr)
            LOG.error('Failed to run: "%s"', ' '.join(clang_version_cmd))
            return False


def has_analyzer_config_option(clang_bin, config_option_name):
    """Check if an analyzer config option is available."""
    cmd = [clang_bin, "-cc1", "-analyzer-config-help"]

    LOG.debug('run: "%s"', ' '.join(cmd))

    try:
        proc = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=analyzer_context.get_context().analyzer_env,
            encoding="utf-8", errors="ignore")
        out, err = proc.communicate()
        LOG.debug("stdout:\n%s", out)
        LOG.debug("stderr:\n%s", err)

        match = re.search(config_option_name, out)
        if match:
            LOG.debug("Config option '%s' is available.", config_option_name)
        return (True if match else False)

    except OSError:
        LOG.error('Failed to run: "%s"', ' '.join(cmd))
        raise


def has_analyzer_option(clang_bin, feature):
    """Test if the analyzer has a specific option.

    Testing a feature is done by compiling a dummy file."""
    with tempfile.NamedTemporaryFile("w", encoding='utf-8') as inputFile:
        inputFile.write("void foo(){}")
        inputFile.flush()
        cmd = [clang_bin, "-x", "c", "--analyze"]
        cmd.extend(feature)
        cmd.extend([inputFile.name, "-o", "-"])

        LOG.debug('run: "%s"', ' '.join(cmd))
        try:
            proc = subprocess.Popen(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                env=analyzer_context.get_context().analyzer_env,
                encoding="utf-8", errors="ignore")
            out, err = proc.communicate()
            LOG.debug("stdout:\n%s", out)
            LOG.debug("stderr:\n%s", err)

            return proc.returncode == 0
        except OSError:
            LOG.error('Failed to run: "%s"', ' '.join(cmd))
            return False



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Argument handlers for the 'CodeChecker cmd permissions' subcommands.
"""

from typing import Dict

from codechecker_api.Authentication_v6.ttypes import AccessControl

from codechecker_common import logger

from .client import init_auth_client
from .cmd_line import CmdLineOutputEncoder
from .product import split_server_url

# Needs to be set in the handler functions.
LOG = None


def init_logger(level, stream=None, logger_name='system'):
    logger.setup_logger(level, stream)
    global LOG
    LOG = logger.get_logger(logger_name)


def __convert_permissions(permissions: AccessControl) -> Dict:
    """ Convert the given permissions to dictionary. """
    ret = {"user_permissions": {}, "group_permissions": {}}

    for user_name, perms in permissions.user.items():
        ret["user_permissions"][user_name] = perms

    for group_name, perms in permissions.group.items():
        ret["group_permissions"][group_name] = perms

    return ret


def handle_permissions(args):
    """
    Argument handler for the 'CodeChecker cmd permissions' subcommand
    """
    # If the given output format is not 'table', redirect logger's output to
    # the stderr.
    stream = None
    if 'output_format' in args:
        stream = 'stderr'

    init_logger(args.verbose if 'verbose' in args else None, stream)

    protocol, host, port = split_server_url(args.server_url)
    client = init_auth_client(protocol, host, port)
    access_control = client.getAccessControl()

    if args.output_format == 'json':
        print(CmdLineOutputEncoder().encode({
            'version': 1,
            'global_permissions': __convert_permissions(
                access_control.globalPermissions),
            'product_permissions': {
                name: __convert_permissions(val)
                for name, val in access_control.productPermissions.items()
            },
        }))



#!/usr/bin/env python3
# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

"""
Run bazel aquery to produce compilation database (compile_commands.json)
for particular bazel build command
"""

from __future__ import print_function
import argparse
import json
import logging
import os
import re
import shlex
import subprocess


def parse_args():
    """
    Parse command line arguments or show help.
    """
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        description=__doc__)
    parser.add_argument("-o", "--output",
                        default="compile_commands.json",
                        help="output compile_commands.json file")
    parser.add_argument("-b", "--build",
                        help="bazel build command arguments")
    parser.add_argument("-v", "--verbosity",
                        default=1,
                        action="count",
                        help="enable verbose logging")
    parser.add_argument("--log-format",
                        default="[bazel] %(levelname)5s: %(message)s",
                        help=argparse.SUPPRESS)

    options = parser.parse_args()

    if options.verbosity >= 2:
        log_level = logging.DEBUG
    elif options.verbosity >= 1:
        log_level = logging.INFO
    else:
        log_level = logging.WARN
    logging.basicConfig(level=log_level, format=options.log_format)

    return options


def split_to_list(arguments):
    """
    Split argument string to list
    """
    if type(arguments) is list:
        return arguments
    else:
        return shlex.split(arguments)


def run_command(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE):
    """
    Run shell command
    """
    cmd = split_to_list(command)
    process = subprocess.Popen(cmd, stdout=stdout, stderr=stderr)
    out, err = process.communicate()
    if process.returncode:
        logging.error("Command: %s...\nError: %s", command, err)
    return out.decode("utf-8")


def bazel_aquery(build_arguments):
    """
    Run bazel query and return list of source files, e.g.:
    bazel cquery 'kind("source file", deps("@workspace//target"))'
        --config=something --noimplicit_deps --notool_deps
    """
    arguments = split_to_list(build_arguments)
    targets = []
    configs = []
    bazel = arguments[0]
    logging.debug("Bazel: %s", bazel)
    logging.debug("Bazel build arguments:")
    for argument in arguments:
        logging.debug("   %s", argument)
        if re.search(r"^(@|:|\/\/)\S*", argument) or argument == "...":
            targets.append(argument)
        elif re.search(r"^--\S*", argument):
            configs.append(argument)
    logging.debug("Targets: %s", targets)
    logging.debug("Configs: %s", configs)
    if len(targets) < 1:
        logging.error(
            "No targets found for bazel query in: %s",
            build_arguments)
        return

    deps = " + ".join([
        "deps(\"{target}\")".format(target=target) for target in targets
    ])
    logging.debug("Deps: %s", deps)

    options = [
        "--output=jsonproto",
        "--include_artifacts=false",
    ]
    function = "mnemonic(\"CppCompile\", {deps})".format(deps=deps)
    command = "{bazel} aquery {options} '{function}' {configs}".format(
        bazel=bazel,
        options=" ".join(options),
        function=function,
        configs=" ".join(configs))
    logging.debug("Command: %s", command)

    out = run_command(command)
    if out:
        data = json.loads(out)
        return data


def get_compile_commands(data,
                         directories=["."],
                         base_dir=None,
                         output_base=None):
    """
    Produce compile commands from action graph JSON data
    """
    compile_commands = []
    not_found = 0
    skipped = 0
    for action in data["actions"]:
        if "mnemonic" in action and action["mnemonic"] in ["CppCompile"]:
            arguments = action["arguments"]
            try:
                index = arguments.index("-c")
                file = arguments[index + 1]
                if not file.endswith(".asm"):
                    (full_path, directory) = find_file(file, directories)
                    if full_path and directory:
                        if base_dir:
                            directory = base_dir
                        compile_command = {
                            "file": full_path,
                            "directory": directory,
                            "command": " ".join(arguments),
                        }
                        filter_compile_command(compile_command, output_base)
                        compile_commands.append(compile_command)
                        logging.info("Add file: %s", file)
                    else:
                        logging.warning("Cannot find: %s", file)
                        not_found += 1
                else:
                    logging.info("Skipping: %s", file)
                    skipped += 1
            except ValueError:
                logging.warning("Wrong compile command: %s", str(arguments))
    logging.info("%d files added, %d skipped (asm), %d not found",
                 len(compile_commands), skipped, not_found)
    return compile_commands


def find_file(file, directories):
    """
    Find file in the list of directories.
    Return full path and corresponding directory.
    """
    for directory in directories:
        full_path = os.path.join(directory, file)
        if os.path.isfile(full_path):
            return (full_path, directory)
    return (False, False)


def filter_compile_command(compile_command, output_base):
    """
    Remove irrelevant options and change paths
    """
    command = compile_command["command"]
    if output_base:
        # bazel-out/k8-fastbuild/bin/external/ -> /abs/path/external/
        command = re.sub(r"( |=)bazel-out\/\S*\/bin\/external\/",
                         r"\1" + output_base + "/external/",
                         command)
    else:
        # bazel-out/k8-fastbuild/bin/external/ -> ../../external/
        command = re.sub(r"( |=)bazel-out\/\S*\/bin\/external\/",
                         r"\1../../external/",
                         command)
    # Remove unsupported options (clang)
    command = re.sub(r" -MD ", " ", command)
    command = re.sub(r" -MF \S* ", " ", command)
    command = re.sub(r" -MT \S* ", " ", command)
    compile_command["command"] = command


def run(build_command, output_file_name):
    """
    Run bazel aquaery and generate compile_commands.json:
    - get bazel info params
    - run bazel aquery and parse output as JSON data
    - create compile commands from JSON data:
      * take only arguments
      * add: file, directory, command
      * change to full paths
      * filter compiler options
    - save to output file (compile_commands.json)
    """
    if not build_command:
        logging.error("No bazel build command specified")
        return False

    if not output_file_name:
        logging.error("No output file is specified")
        return False

    bazel_info = split_to_list(build_command)[0] + " info "
    bazel_workspace = run_command(bazel_info + "workspace").rstrip()
    bazel_output_base = run_command(bazel_info + "output_base").rstrip()
    bazel_execution_root = run_command(bazel_info + "execution_root").rstrip()

    bazel_work_dir = bazel_execution_root.replace("/execroot/", "/external/")
    if not os.path.exists(bazel_work_dir):
        bazel_work_dir = bazel_workspace

    logging.info("Bazel workspace:     %s", bazel_workspace)
    logging.info("Bazel output base:   %s", bazel_output_base)
    logging.info("Bazel work dir:      %s", bazel_work_dir)

    logging.info("Bazel build options: %s", build_command)
    data = bazel_aquery(build_command)
    if not data:
        logging.error("Command '%s'", build_command)
        logging.error("Produces no output")
        return False

    directories = [
        bazel_output_base,
        bazel_work_dir,
    ]
    compile_commands = get_compile_commands(
        data, directories, output_base=bazel_output_base)

    logging.info("Saving to: %s", output_file_name)
    with open(output_file_name, "w") as compile_commands_file:
        json.dump(compile_commands, compile_commands_file, indent=4)
    return True


def main():
    """
    Main function
    """
    options = parse_args()
    logging.debug("Options: %s", options)
    if not run(options.build, options.output):
        exit(1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Converts Gerrit Review changed files list to CodeChecker skipfile.
"""

import argparse
import json
import re


def create_skipfile(files_changed, skipfile):
    # File is likely to contain some garbage values at start,
    # only the corresponding json should be parsed.
    json_pattern = re.compile(r"^\{.*\}")
    for line in files_changed.readlines():
        if re.match(json_pattern, line):
            for filename in json.loads(line):
                if "/COMMIT_MSG" in filename:
                    continue
                skipfile.write("+*/%s\n" % filename)

    skipfile.write("-*\n")


def main():
    parser = argparse.ArgumentParser(
        description="Converts Gerrit Review changed files "
        "json to CodeChecker skipfile."
    )
    parser.add_argument(
        "files_changed",
        type=argparse.FileType("r"),
        help="Path of changed files json from Gerrit.",
    )
    parser.add_argument(
        "skipfile",
        nargs="?",
        default="skipfile",
        type=argparse.FileType("w"),
        help="Path of the skipfile output. Default is ./skipfile.",
    )
    args = parser.parse_args()

    create_skipfile(args.files_changed, args.skipfile)


if __name__ == "__main__":
    main()



"""
Add CodeChecker version to run_history

Revision ID: 9987aa593ca7
Revises:     e89887e7d3f0
Create Date: 2018-09-05 17:43:42.099167
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = '9987aa593ca7'
down_revision = 'e89887e7d3f0'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('run_histories',
                  sa.Column('cc_version', sa.String(), nullable=True))


def downgrade():
    op.drop_column('run_histories', 'client_version')



"""
Report annotations

Revision ID: 9d956a0fae8d
Revises:     75ae226b5d88
Create Date: 2023-02-09 17:45:56.162040
"""

from alembic import op
import sqlalchemy as sa

# Revision identifiers, used by Alembic.
revision = '9d956a0fae8d'
down_revision = '75ae226b5d88'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'report_annotations',
        sa.Column('report_id', sa.Integer(), nullable=False),
        sa.Column('key', sa.String(), nullable=False),
        sa.Column('value', sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ['report_id'], ['reports.id'],
            name=op.f('fk_report_annotations_report_id_reports'),
            ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('report_id', 'key',
                                name=op.f('pk_report_annotations'))
    )
    op.create_index(op.f('ix_report_annotations_report_id'),
                    'report_annotations', ['report_id'], unique=False)


def downgrade():
    op.drop_index(op.f('ix_report_annotations_report_id'),
                  table_name='report_annotations')
    op.drop_table('report_annotations')



"""
Add index for report and history id columns

Revision ID: a24461972d2e
Revises:     dabc6998b8f0
Create Date: 2021-06-10 15:38:59.504534
"""

from alembic import op


# Revision identifiers, used by Alembic.
revision = 'a24461972d2e'
down_revision = 'dabc6998b8f0'
branch_labels = None
depends_on = None


def upgrade():
    op.create_index(
        op.f('ix_report_analysis_info_report_id'),
        'report_analysis_info',
        ['report_id'],
        unique=False)

    op.create_index(
        op.f('ix_run_history_analysis_info_run_history_id'),
        'run_history_analysis_info',
        ['run_history_id'],
        unique=False)


def downgrade():
    op.drop_index(
        op.f('ix_run_history_analysis_info_run_history_id'),
        table_name='run_history_analysis_info')

    op.drop_index(
        op.f('ix_report_analysis_info_report_id'),
        table_name='report_analysis_info')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Functions to check the host machine and the analyzers for various
features, dependecies and configurations.
"""


import os

from codechecker_common.logger import get_logger

LOG = get_logger('system')


def check_zlib():
    """ Check if zlib compression is available.
    If wrong libraries are installed on the host machine it is
    possible the the compression fails which is required to
    store data into the database.
    """
    try:
        import zlib
        zlib.compress(b'Compress this')
        return True
    except Exception as ex:
        LOG.error(str(ex))
        LOG.error('Failed to import zlib module.')
        return False


def get_postgresql_driver_name():
    # pylint: disable=unused-variable
    try:
        driver = os.getenv('CODECHECKER_DB_DRIVER')
        if driver:
            return driver

        try:
            # pylint: disable=W0611
            import psycopg2
            return "psycopg2"
        except Exception:
            # pylint: disable=W0611
            import pg8000
            return "pg8000"
    except Exception as ex:
        LOG.error(str(ex))
        LOG.error('Failed to import psycopg2 or pg8000 module.')
        raise


def check_postgresql_driver():
    try:
        get_postgresql_driver_name()
        return True
    except Exception as ex:
        LOG.debug(ex)
        return False


def check_sql_driver(check_postgresql):
    # pylint: disable=unused-variable
    if check_postgresql:
        try:
            get_postgresql_driver_name()
            return True
        except Exception:
            return False
    else:
        try:
            try:
                # pylint: disable=W0611
                import pysqlite2
            except Exception:
                # pylint: disable=W0611
                import sqlite3
        except Exception as ex:
            LOG.debug(ex)
            return False
        return True



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import logging
import os
import json

from typing import Dict, List

from codechecker_report_converter.report import File, get_or_create_file, \
    Report

from ..analyzer_result import AnalyzerResultBase


LOG = logging.getLogger('report-converter')


class AnalyzerResult(AnalyzerResultBase):
    """ Transform analyzer result of the TSLint analyzer. """

    TOOL_NAME = 'tslint'
    NAME = 'TSLint'
    URL = 'https://palantir.github.io/tslint'

    def get_reports(self, result_file_path: str) -> List[Report]:
        """ Parse the given analyzer result. """
        reports: List[Report] = []

        if not os.path.exists(result_file_path):
            LOG.error("Report file does not exist: %s", result_file_path)
            return reports

        try:
            with open(result_file_path, 'r',
                      encoding="utf-8", errors="ignore") as report_f:
                bugs = json.load(report_f)
        except (IOError, json.decoder.JSONDecodeError):
            LOG.error("Failed to parse the given analyzer result '%s'. Please "
                      "give a valid json file generated by TSLint.",
                      result_file_path)
            return reports

        file_cache: Dict[str, File] = {}
        for bug in bugs:
            file_path = os.path.join(
                os.path.dirname(result_file_path), bug.get('name'))

            if not os.path.exists(file_path):
                LOG.warning("Source file does not exists: %s", file_path)
                continue

            end_pos = bug['startPosition']
            line = int(end_pos['line'] + 1)
            col = int(end_pos['character'] + 1)

            reports.append(Report(
                get_or_create_file(os.path.abspath(file_path), file_cache),
                line, col, bug['failure'], bug['ruleName']
            ))

        return reports



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import json
import logging
import os

from typing import Dict, List

from codechecker_report_converter.report import File, get_or_create_file, \
    Report

from ..analyzer_result import AnalyzerResultBase


LOG = logging.getLogger('report-converter')


class AnalyzerResult(AnalyzerResultBase):
    """ Transform analyzer result of the Pylint analyzer. """

    TOOL_NAME = 'pylint'
    NAME = 'Pylint'
    URL = 'https://www.pylint.org'

    def get_reports(self, result_file_path: str) -> List[Report]:
        """ Get reports from the given analyzer result. """
        reports: List[Report] = []

        if not os.path.exists(result_file_path):
            LOG.error("Report file does not exist: %s", result_file_path)
            return reports

        try:
            with open(result_file_path, 'r',
                      encoding="utf-8", errors="ignore") as f:
                bugs = json.load(f)
        except (IOError, json.decoder.JSONDecodeError):
            LOG.error("Failed to parse the given analyzer result '%s'. Please "
                      "give a valid json file generated by Pylint.",
                      result_file_path)
            return reports

        file_cache: Dict[str, File] = {}
        for bug in bugs:
            file_path = os.path.join(
                os.path.dirname(result_file_path), bug.get('path'))
            if not os.path.exists(file_path):
                LOG.warning("Source file does not exists: %s", file_path)
                continue

            reports.append(Report(
                get_or_create_file(os.path.abspath(file_path), file_cache),
                int(bug['line']),
                int(bug['column']),
                bug['message'],
                bug['symbol']))

        return reports



"""
Cleanup plan

Revision ID: fb356f0eefed
Revises:     ad2a567e513a
Create Date: 2021-09-06 10:55:43.093729
"""

from alembic import op
import sqlalchemy as sa


# Revision identifiers, used by Alembic.
revision = 'fb356f0eefed'
down_revision = 'ad2a567e513a'
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'cleanup_plans',
        sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
        sa.Column('name', sa.String(), nullable=False),
        sa.Column('due_date', sa.DateTime(), nullable=True),
        sa.Column('description', sa.String(), nullable=True),
        sa.Column('closed_at', sa.DateTime(), nullable=True),
        sa.PrimaryKeyConstraint('id', name=op.f('pk_cleanup_plans')),
        sa.UniqueConstraint('name', name=op.f('uq_cleanup_plans_name'))
    )

    op.create_table(
        'cleanup_plan_report_hashes',
        sa.Column('cleanup_plan_id', sa.Integer(), nullable=True),
        sa.Column('bug_hash', sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ['cleanup_plan_id'],
            ['cleanup_plans.id'],
            name=op.f('fk_cleanup_plan_report_hashes_cleanup_plan_id_'
                      'cleanup_plans'),
            ondelete='CASCADE',
            initially='DEFERRED',
            deferrable=True),
        sa.PrimaryKeyConstraint(
            'bug_hash',
            name=op.f('pk_cleanup_plan_report_hashes'))
    )

    op.create_index(
        op.f('ix_cleanup_plan_report_hashes_cleanup_plan_id'),
        'cleanup_plan_report_hashes',
        ['cleanup_plan_id'],
        unique=False)


def downgrade():
    op.drop_index(
        op.f('ix_cleanup_plan_report_hashes_cleanup_plan_id'),
        table_name='cleanup_plan_report_hashes')

    op.drop_table('cleanup_plan_report_hashes')
    op.drop_table('cleanup_plans')



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

"""
Post-process statistical results.

Clang output files will be parsed for outputs by the statistics collector
checkers and converted into a special yml file which can be parsed back by the
statistics checkers.
"""

import logging
import os

from .collectors.return_value import ReturnValueCollector
from .collectors.special_return_value import SpecialReturnValueCollector

LOG = logging.getLogger('StatisticsCollector')


def process(input_dir, output_dir,
            stats_min_sample_count, stats_relevance_threshold):
    """
    Read the clang analyzer outputs where the statistics emitter checkers
    were enabled and collect the statistics.

    After the statistics collection cleanup the output files.
    """

    try:
        os.stat(output_dir)
    except Exception as ex:
        LOG.debug(ex)
        os.mkdir(output_dir)

    if not os.path.exists(input_dir):
        LOG.debug("No statistics directory was found")
        return

    clang_outs = []
    try:
        for f in os.listdir(input_dir):
            if os.path.isfile(os.path.join(input_dir, f)):
                clang_outs.append(os.path.join(input_dir, f))
    except OSError as oerr:
        LOG.debug(oerr)
        LOG.debug("Statistics can not be collected.")
        LOG.debug("Analyzer output error.")
        return

    if not clang_outs:
        LOG.warning("No output files were found to collect statistics.")
        return

    ret_collector = ReturnValueCollector(stats_min_sample_count,
                                         stats_relevance_threshold)
    special_ret_collector =\
        SpecialReturnValueCollector(stats_min_sample_count,
                                    stats_relevance_threshold)

    lines = set()
    for clang_output in clang_outs:
        with open(clang_output, 'r',
                  encoding='utf-8', errors='ignore') as out:
            lines |= set(out.readlines())
    for line in lines:
        ret_collector.process_line(line)
        special_ret_collector.process_line(line)
    LOG.debug("Collecting statistics finished.")

    # Write out statistics.
    unchecked_yaml = ReturnValueCollector.stats_file(output_dir)
    LOG.debug("Writing out statistics to %s", unchecked_yaml)
    with open(unchecked_yaml, 'w',
              encoding='utf-8', errors='ignore') as uyaml:
        uyaml.write(ret_collector.get_yaml())

    special_ret_yaml = SpecialReturnValueCollector.stats_file(output_dir)
    LOG.debug("Writing out statistics to %s", special_ret_yaml)
    with open(special_ret_yaml, 'w',
              encoding='utf-8', errors='ignore') as uyaml:
        uyaml.write(special_ret_collector.get_yaml())



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
List the checkers available in the analyzers.
"""


import argparse
import json
import os
import sys
from collections import defaultdict
from typing import Dict, Iterable, List, Tuple

from codechecker_report_converter import twodim

from codechecker_analyzer import analyzer_context
from codechecker_analyzer.analyzers import analyzer_types

from codechecker_common import arg, logger
from codechecker_common.output import USER_FORMATS
from codechecker_common.checker_labels import CheckerLabels
from codechecker_analyzer.analyzers.config_handler import CheckerState

LOG = logger.get_logger('system')


def get_argparser_ctor_args():
    """
    This method returns a dict containing the kwargs for constructing an
    argparse.ArgumentParser (either directly or as a subparser).
    """

    data_files_dir_path = analyzer_context.get_context().data_files_dir_path
    labels_dir_path = os.path.join(data_files_dir_path, 'config', 'labels')
    return {
        'prog': 'CodeChecker checkers',
        'formatter_class': arg.RawDescriptionDefaultHelpFormatter,

        # Description is shown when the command's help is queried directly
        'description': "Get the list of checkers available and their enabled "
                       "status in the supported analyzers.",

        # Epilogue is shown after the arguments when the help is queried
        # directly.
        'epilog': """
The list of checkers that are enabled or disabled by default can be edited by
editing "profile:default" labels in the directory '{}'.

Example scenario: List checkers by labels
-----------------------------------------
List checkers in "sensitive" profile:
    CodeChecker checkers --label profile:sensitive
    CodeChecker checkers --profile sensitive

List checkers in "HIGH" severity:
    CodeChecker checkers --label severity:HIGH
    CodeChecker checkers --severity HIGH

List checkers covering str34-c SEI-CERT rule:
    CodeChecker checkers --label sei-cert:str-34-c
    CodeChecker checkers --guideline sei-cert:str34-c

List checkers covering all SEI-CERT rules:
    CodeChecker checkers --label guideline:sei-cert
    CodeChecker checkers --guideline sei-cert

List available profiles, guidelines and severities:
    CodeChecker checkers --profile
    CodeChecker checkers --guideline
    CodeChecker checkers --severity

List labels and their available values:
    CodeChecker checkers --label
    CodeChecker checkers --label severity
""".format(os.path.join(labels_dir_path)),

        # Help is shown when the "parent" CodeChecker command lists the
        # individual subcommands.
        'help': "List the checkers available for code analysis."
    }


def add_arguments_to_parser(parser):
    """
    Add the subcommand's arguments to the given argparse.ArgumentParser.
    """

    parser.add_argument('--analyzers',
                        nargs='+',
                        dest='analyzers',
                        metavar='ANALYZER',
                        required=False,
                        choices=analyzer_types.supported_analyzers,
                        default=list(analyzer_types.supported_analyzers.
                                     keys()),
                        help="Show checkers only from the analyzers "
                             "specified.")

    parser.add_argument('-w', '--warnings',
                        dest='show_warnings',
                        default=argparse.SUPPRESS,
                        action='store_true',
                        required=False,
                        help="DEPRECATED. Show available warning flags.")

    parser.add_argument('--details',
                        dest='details',
                        default=argparse.SUPPRESS,
                        action='store_true',
                        required=False,
                        help="Show details about the checker, such as "
                             "status, checker name, analyzer name, severity, "
                             "guidelines and description. Status shows if the "
                             "checker is enabled besides the given labels. "
                             "If the labels don't trigger a checker then the "
                             "status is determined by the analyzer tool.")

    parser.add_argument('--label',
                        nargs='?',
                        required=False,
                        default=argparse.SUPPRESS,
                        help="Filter checkers that are attached the given "
                             "label. The format of a label is "
                             "<label>:<value>. If no argument is given then "
                             "available labels are listed. If only <label> is "
                             "given then available values are listed.")

    parser.add_argument('--profile',
                        dest='profile',
                        nargs='?',
                        required=False,
                        default=argparse.SUPPRESS,
                        help="List checkers enabled by the selected profile. "
                             "If no argument is given then available profiles "
                             "are listed.")

    parser.add_argument('--guideline',
                        dest='guideline',
                        nargs='?',
                        required=False,
                        default=argparse.SUPPRESS,
                        help="List checkers that report on a specific "
                             "guideline. Without additional parameter, the "
                             "available guidelines and their corresponding "
                             "rules will be listed.")

    parser.add_argument('--severity',
                        dest='severity',
                        nargs='?',
                        required=False,
                        default=argparse.SUPPRESS,
                        help="List checkers with the given severity. Make "
                             "sure to indicate severity in capitals (e.g. "
                             "HIGH, MEDIUM, etc.) If no argument is given "
                             "then available severities are listed.")

    parser.add_argument('--checker-config',
                        dest='checker_config',
                        default=argparse.SUPPRESS,
                        action='store_true',
                        required=False,
                        help="Show checker configuration options for all "
                             "existing checkers supported by the analyzer. "
                             "These can be given to 'CodeChecker analyze "
                             "--checker-config'.")

    parser.add_argument('-o', '--output',
                        dest='output_format',
                        required=False,
                        default='custom',
                        choices=USER_FORMATS + ['custom'],
                        help="The format to list the applicable checkers as.")

    logger.add_verbose_arguments(parser)
    parser.set_defaults(func=main)


def __uglify(text: str) -> str:
    """
    csv and json format output contain this non human readable header string:
    no CamelCase and no space.
    """
    return text.lower().replace(' ', '_')


def __guideline_to_label(
    args: argparse.Namespace,
    cl: CheckerLabels
) -> str:
    """
    Transforms --guideline parameter as if they were given through --label.
    For example "--guideline sei-cert" is equivalent with
    "--label guideline:sei-cert" and "--guideline sei-cert:str38-c" is the same
    as "--label sei-cert:str38-c".
    """
    guidelines = []
    for analyzer in args.analyzers:
        guidelines.extend(cl.occurring_values('guideline', analyzer))

    if args.guideline in guidelines:
        return f'guideline:{args.guideline}'
    else:
        if args.guideline.find(':') == -1:
            LOG.error('--guideline parameter is either <guideline> or '
                      '<guideline>:<rule>')
            sys.exit(1)
        return args.guideline


def __get_detailed_checker_info(
    args: argparse.Namespace,
    cl: CheckerLabels
) -> Dict[str, list]:
    """
    Returns a dictionary which maps analyzer names to the collection of their
    supported checkers. Checker information is described with tuples of this
    information: (status, checker name, analyzer name, description, labels).
    """
    working_analyzers, _ = analyzer_types.check_supported_analyzers(
        analyzer_types.supported_analyzers)

    analyzer_config_map = analyzer_types.build_config_handlers(
        args, working_analyzers)

    checker_info = defaultdict(list)

    for analyzer in working_analyzers:
        config_handler = analyzer_config_map.get(analyzer)
        analyzer_class = analyzer_types.supported_analyzers[analyzer]

        checkers = analyzer_class.get_analyzer_checkers()

        profile_checkers = []
        if 'profile' in args:
            available_profiles = cl.get_description('profile')

            if args.profile not in available_profiles:
                LOG.error("Checker profile '%s' does not exist!",
                          args.profile)
                LOG.error("To list available profiles, use '--profile list'.")
                sys.exit(1)

            profile_checkers.append((f'profile:{args.profile}', True))

        if 'label' in args:
            profile_checkers.extend((label, True) for label in args.label)

        if 'severity' in args:
            profile_checkers.append((f'severity:{args.severity}', True))

        if 'guideline' in args:
            profile_checkers.append((__guideline_to_label(args, cl), True))

        config_handler.initialize_checkers(checkers, profile_checkers)

        for checker, (state, description) in config_handler.checks().items():
            labels = cl.labels_of_checker(checker, analyzer)
            state = CheckerState.enabled if ('profile', 'default') in labels \
                else CheckerState.disabled
            checker_info[analyzer].append(
                (state, checker, analyzer, description, sorted(labels)))

    return checker_info


def __print_profiles(args: argparse.Namespace, cl: CheckerLabels):
    """
    Print checker profiles according to the command line arguments to the
    standard output.
    """
    if args.output_format == 'custom':
        args.output_format = 'rows'

    if 'details' in args:
        header = ['Profile name', 'Description']
        rows = cl.get_description('profile').items()
    else:
        header = ['Profile name']
        rows = [(key,) for key in cl.get_description('profile')]

    if args.output_format in ['csv', 'json']:
        header = list(map(__uglify, header))

    print(twodim.to_str(args.output_format, header, rows))


def __print_severities(args: argparse.Namespace, cl: CheckerLabels):
    """
    Print checker severities according to the command line arguments to the
    standard output.
    """
    if args.output_format == 'custom':
        args.output_format = 'rows'

    if 'details' in args:
        header = ['Severity', 'Description']
        rows = cl.get_description('severity').items()
    else:
        header = ['Severity']
        rows = [(key,) for key in cl.get_description('severity')]

    if args.output_format in ['csv', 'json']:
        header = list(map(__uglify, header))

    print(twodim.to_str(args.output_format, header, rows))


def __print_guidelines(args: argparse.Namespace, cl: CheckerLabels):
    """
    Print guidelines according to the command line arguments to the standard
    output.
    """
    if args.output_format == 'custom':
        args.output_format = 'rows'

    result = {}

    for guideline in cl.get_description('guideline'):
        result[guideline] = set(cl.occurring_values(guideline))

    header = ['Guideline', 'Rules']
    if args.output_format in ['csv', 'json']:
        header = list(map(__uglify, header))

    if args.output_format == 'json':
        rows = [(g, sorted(list(r))) for g, r in result.items()]
    else:
        rows = [(g, ', '.join(sorted(r))) for g, r in result.items()]

    if args.output_format == 'rows':
        for row in rows:
            print('Guideline: {}'.format(row[0]))
            print('Rules: {}'.format(row[1]))
    else:
        print(twodim.to_str(args.output_format, header, rows))


def __print_labels(args: argparse.Namespace, cl: CheckerLabels):
    """
    Print labels according to the command line arguments to the standard
    output.
    """
    if args.output_format == 'custom':
        args.output_format = 'rows'

    header = ['Label']
    if args.output_format in ['csv', 'json']:
        header = list(map(__uglify, header))

    rows = list(map(lambda x: (x,), cl.labels()))

    print(twodim.to_str(args.output_format, header, rows))


def __print_label_values(args: argparse.Namespace, cl: CheckerLabels):
    """
    If --label flag is given an argument which doesn't contain a colon (:) then
    we assume that the user intends to see the available values of that label:
    CodeChecker checkers --label severity
    """
    if args.output_format == 'custom':
        args.output_format = 'rows'

    header = ['Value']
    if args.output_format == 'custom':
        args.output_format = 'rows'

    rows = list(map(lambda x: (x,), cl.occurring_values(args.label)))

    if rows:
        print(twodim.to_str(args.output_format, header, rows))
    else:
        LOG.info(
            'Label "%s" doesn\'t exist. Use "CodeChecker checkers --label" '
            'command to list available labels.', args.label)


def __format_row(row: Tuple) -> Tuple:
    """
    Perform some formatting of the detailed checker info.
    row -- A tuple with detailed checker info coming from
           __get_detailed_checker_info() function.
    """
    state = '+' if row[0] == CheckerState.enabled else '-'
    labels = ', '.join(f'{k}:{v}' for k, v in row[4])

    return state, row[1], row[2], row[3], labels


def __print_checkers_custom_format(checkers: Iterable):
    """
    This function prints a detailed view of the selected checkers in a custom
    format. Due to its customness it isn't implemented in module twodim.
    """
    for checker in checkers:
        if checker[0] == CheckerState.enabled:
            status = 'enabled'
        elif checker[0] == CheckerState.disabled:
            status = 'disabled'

        print(checker[1])
        print('  Status:', status)
        print('  Analyzer:', checker[2])
        print('  Description:', checker[3])
        print('  Labels:')

        for label in checker[4]:
            print(f'    {label[0]}:{label[1]}')

        print()


def __print_checkers_json_format(checkers: Iterable, detailed: bool):
    """
    Print checker information in JSON format. This function is implemented to
    shorten __print_checkers(). JSON format is handled separately because its
    structure differs from other twodim formats.
    """
    def checker_info_dict(c):
        if c[0] == CheckerState.enabled:
            status = 'enabled'
        elif c[0] == CheckerState.disabled:
            status = 'disabled'
        else:
            status = 'unknown'

        return {
            'status': status,
            'name': c[1],
            'analyzer': c[2],
            'description': c[3],
            'labels': list(map(lambda x: f'{x[0]}:{x[1]}', c[4]))}

    if detailed:
        print(json.dumps([checker_info_dict(c) for c in checkers]))
    else:
        print(json.dumps([c[1] for c in checkers]))


def __post_process_result(result: List[Tuple]):
    """ Postprocess the given result.

    It will update the value of the doc_url label and create an absolute file
    path if it is a relative path.
    """
    data_files_dir_path = os.environ.get('CC_DATA_FILES_DIR', '')
    www_dir_path = os.path.join(data_files_dir_path, 'www')
    for res in result:
        for idx, (name, value) in enumerate(res[4]):
            if name == 'doc_url' and not value.startswith('http'):
                res[4][idx] = (name, os.path.normpath(
                    os.path.join(www_dir_path, value.strip(os.sep))))


def __print_checkers(args: argparse.Namespace, cl: CheckerLabels):
    """
    Print checkers according to the command line arguments to the standard
    output.
    """
    labels = [args.label] if 'label' in args else []

    if 'profile' in args:
        labels.append(f'profile:{args.profile}')

    if 'guideline' in args:
        labels.append(__guideline_to_label(args, cl))

    if 'severity' in args:
        labels.append(f'severity:{args.severity}')

    checker_info = __get_detailed_checker_info(args, cl)

    result = []
    for analyzer in args.analyzers:
        if labels:
            checkers = cl.checkers_by_labels(labels, analyzer)
            result.extend(
                filter(lambda x: x[1] in checkers, checker_info[analyzer]))
        else:
            result.extend(checker_info[analyzer])

    __post_process_result(result)

    if args.output_format == 'custom':
        if result:
            __print_checkers_custom_format(result)
        else:
            LOG.info('No checkers with the given label found.')
        return

    if args.output_format == 'json':
        __print_checkers_json_format(result, 'details' in args)
        return

    if 'details' in args:
        header = ['Status', 'Name', 'Analyzer', 'Description', 'Labels']
        rows = list(map(__format_row, result))
    else:
        header = ['Name']
        rows = [[r[1]] for r in result]

    if args.output_format in ['csv', 'json']:
        header = list(map(__uglify, header))

    if rows:
        print(twodim.to_str(args.output_format, header, rows))
    else:
        LOG.info('No checkers with the given label found.')


def __print_checker_config(args: argparse.Namespace):
    """
    Print checker config options according to the command line arguments to the
    standard output. The set of config options comes from the analyzers.
    """
    if args.output_format == 'custom':
        args.output_format = 'rows'

    working_analyzers, errored = analyzer_types.check_supported_analyzers(
        args.analyzers)
    analyzer_types.check_available_analyzers(working_analyzers, errored)

    if 'details' in args:
        header = ['Option', 'Description']
    else:
        header = ['Option']

    if args.output_format in ['csv', 'json']:
        header = list(map(__uglify, header))

    rows = []
    analyzer_failures = []
    for analyzer in working_analyzers:
        analyzer_class = analyzer_types.supported_analyzers[analyzer]

        configs = analyzer_class.get_checker_config()
        if not configs:
            # Checker configurations are not supported by cppcheck
            if analyzer != "cppcheck":
                analyzer_failures.append(analyzer)
                continue
            # TODO
            if analyzer != "gcc":
                analyzer_failures.append(analyzer)
                continue

        rows.extend((':'.join((analyzer, c[0])), c[1]) if 'details' in args
                    else (':'.join((analyzer, c[0])),) for c in configs)

    if rows:
        print(twodim.to_str(args.output_format, header, rows))

    analyzer_types.print_unsupported_analyzers(errored)

    if analyzer_failures:
        LOG.error("Failed to get checker configuration options for '%s' "
                  "analyzer(s)! Please try to upgrade your analyzer "
                  "version to use this feature.",
                  ', '.join(analyzer_failures))
        sys.exit(1)


def main(args):
    """
    List the checkers available in the specified (or all supported) analyzers
    alongside with their description or enabled status in various formats.
    """
    # If the given output format is not 'table', redirect logger's output to
    # the stderr.
    logger.setup_logger(args.verbose if 'verbose' in args else None,
                        None if args.output_format == 'table' else 'stderr')

    cl = analyzer_context.get_context().checker_labels

    if 'profile' in args and not args.profile:
        __print_profiles(args, cl)
        return

    if 'severity' in args and not args.severity:
        __print_severities(args, cl)
        return

    if 'guideline' in args and not args.guideline:
        __print_guidelines(args, cl)
        return

    if 'label' in args and not args.label:
        __print_labels(args, cl)
        return

    if 'label' in args and ':' not in args.label:
        __print_label_values(args, cl)
        return

    if 'checker_config' in args:
        __print_checker_config(args)
        return

    __print_checkers(args, cl)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import logging
import os

from typing import Dict, Iterator, List, Optional, Tuple

from codechecker_report_converter.report import File, Report
from codechecker_report_converter.report.checker_labels import CheckerLabels
from codechecker_report_converter.report.hash import HashType
from codechecker_report_converter.report.parser import plist, sarif
from codechecker_report_converter.report.parser.base import AnalyzerInfo


LOG = logging.getLogger('report-converter')


SUPPORTED_ANALYZER_TYPES = tuple(sorted([plist.EXTENSION, sarif.EXTENSION]))


SUPPORTED_ANALYZER_EXTENSIONS = \
    tuple([f".{ext}" for ext in SUPPORTED_ANALYZER_TYPES])


def is_supported(analyzer_result_file_path: str) -> bool:
    """ True if the given report file can be parsed. """
    return analyzer_result_file_path.endswith(SUPPORTED_ANALYZER_EXTENSIONS)


def get_parser(
    analyzer_result_file_path: str,
    checker_labels: Optional[CheckerLabels] = None,
    file_cache: Optional[Dict[str, File]] = None
):
    """ Returns a parser object for the given analyzer result file. """
    # FIXME: It would be more elegant to collect these modules (plist, sarif,
    # etc) into a list, but mypy seems to not recognize attributes from modules
    # saved into variables.
    if analyzer_result_file_path.endswith(plist.EXTENSION):
        return plist.Parser(checker_labels, file_cache)
    if analyzer_result_file_path.endswith(sarif.EXTENSION):
        return sarif.Parser(checker_labels, file_cache)


def get_reports(
    analyzer_result_file_path: str,
    checker_labels: Optional[CheckerLabels] = None,
    file_cache: Optional[Dict[str, File]] = None,
    source_dir_path: Optional[str] = None
) -> List[Report]:
    """ Get reports from the given report file. """
    parser = get_parser(analyzer_result_file_path, checker_labels, file_cache)

    if parser:
        return parser.get_reports(analyzer_result_file_path, source_dir_path)
    else:
        LOG.error(f"Found no parsers to parse {analyzer_result_file_path}! "
                  "Supported file extension types are "
                  f"{SUPPORTED_ANALYZER_EXTENSIONS}.")

    return []


def create(
    output_file_path: str,
    reports: List[Report],
    checker_labels: Optional[CheckerLabels] = None,
    analyzer_info: Optional[AnalyzerInfo] = None
):
    """ Creates an analyzer output file from the given reports. """
    parser = get_parser(output_file_path, checker_labels)

    if parser:
        data = parser.convert(reports, analyzer_info)
        parser.write(data, output_file_path)


def replace_report_hash(
    analyzer_result_file_path: str,
    hash_type=HashType.CONTEXT_FREE
):
    """ Override hash in the given file by using the given version hash. """
    parser = get_parser(analyzer_result_file_path)

    if parser:
        parser.replace_report_hash(
            analyzer_result_file_path, hash_type)


def analyzer_result_files(
    input_paths: List[str]
) -> Iterator[Tuple[str, List[str]]]:
    """
    Iterate over the input paths and returns an iterator of the supported
    analyzer result file paths and metadata information if available.
    """
    for input_path in input_paths:
        input_path = os.path.abspath(input_path)
        LOG.debug("Parsing input argument: '%s'", input_path)

        if os.path.isfile(input_path):
            input_dir_path = os.path.dirname(input_path)
            if is_supported(input_path):
                yield input_dir_path, [input_path]
        elif os.path.isdir(input_path):
            input_dir_path = input_path
            for root_dir, _, file_names in os.walk(input_path):
                analyzer_result_file_paths: List[str] = []
                for file_name in file_names:
                    input_file_path = os.path.join(root_dir, file_name)
                    if is_supported(input_file_path):
                        analyzer_result_file_paths.append(input_file_path)

                yield root_dir, analyzer_result_file_paths
